%{
  //---------------------------------------------------------------------------
  //
  //                     XMM Science Analysis System (SAS)
  //                   (c) 1997-2010 European Space Agency
  //
  //---------------------------------------------------------------------------
  //
  //  DssGrammar.y --- grammar for DSS expressions
  //
  //  DESCRIPTION:
  //      This is a yacc/bison input file which defines the grammar of
  //      DSS-compliant selection expressions.
  //
  //  CHANGE HISTORY:
  //      UL, 1999-11-16: 1.0.0: - first version
  //      UL, 1999-12-17: 1.1.0: - brought in line with selectlib-4.23
  //                               (Region->GeoRegion)
  //
  //  AUTHOR:
  //      U. Lammers (UL), Tue Nov 16 10:35:01 MET 1999
  //
  //  $Header: /data/xmm12/cvsadmin/soc_repos/task_devel/packages/dsslib/src/dssbuild/DssParser.y,v 1.3 1999/12/17 10:51:36 ulammers Exp $
  ///

#include <cmath>
#include <climits>
#include <string>
#include <stdio.h>

#include "basicdefs.h"
#include "Dal.h"
#include "DssBuilder.h"

#define YYDEBUG 1

#define yylex       dsslex
  extern int dsslex();        //  this is the name of the lexer function
                              //  that flex generates with the -Pdss
                              //  option; we do need a distinct name
                              //  in order to avoid name clashes with
                              //  other lexers in the SAS

  namespace DssParser {

    extern void yyerror(char *);    //  The locally added functions in the
                                    //  lexer are in a namespace DssParser

    enum RelOp {LtOpT=1, LeOpT, GtOpT, GeOpT, EqOpT, NeOpT};
    
    extern const Table * yyFilteredTable;

    %}

  %union {
    double                  yNumber;
    string                  *yBlockSpec;
    string                  *yAxisName;
    double                  (*yFunc1)(double);
    double                  (*yFunc2)(double, double);
    DRange                  *yRange;
    DRangeList              *yRangeList;
    vector<float>           *yPointList;
    RelOp                   yRelOp;
  }

    %token PLUS, MINUS, MULT, DIV, EXPO, MODULO
    %token ANDIF, ORIF, NOT, BLOCKFILTER1, BLOCKFILTER2, IN, RELOP
    %token AND, OR, XOR, LSHIFT, RSHIFT, BITCOMPL, TRUECONST, FALSECONST
    %token RPOINT, RLINE, RCIRCLE, RRING, RELLIPSE, RELLIPSERING, RBOX
    %token RRECTANGLE, RPOLYGON, RSECTOR, RDIAMOND
    %token <yBlockSpec> BLOCK
    %token <yNumber>    NUMBER
    %token <yAxisName>  AXISNAME
    %token <yFunc1>     AFUNC1
    %token <yFunc2>     AFUNC2
    %type <yNumber>     arith
    %type <yAxisName>   axisname
    %type <yRange>      intervalspec
    %type <yRangeList>  intervallist
    %type <yPointList>  pointlist
    %type <yRelOp>      RELOP


       /* --( logical/arithmetic operators; incresing precedence )--*/
    %left   ORIF               /* ||                           */
    %left   ANDIF              /* &&                           */
    %left   OR                 /* |                            */
    %left   XOR                /* ^                            */
    %left   AND                /* &                            */
    %left   RELOP              /* comparisons                  */
    %left   LSHIFT RSHIFT      /* <<, >>                       */
    %left   PLUS MINUS         /* plus, minus                  */
    %left   MULT DIV MODULO    /* multiplication, division     */
    %left   EXPO               /* exponentiation               */
    %right  NOT BITCOMPL       /* sign, negation, etc.         */

    %%

dssexpr : dss 


dss:      '(' dss ')'
        | dss ANDIF dss  { dssBuilder.andDss(); }
        | dss ORIF dss   { dssBuilder.orDss(); }
	| filter         { dssBuilder.newDss(); }

axisname: AXISNAME         { $$ = $1; }
        | '(' AXISNAME ')' { $$ = $2; }

filter: axisname IN intervallist
                                { if(yyFilteredTable){
                                    const Column *ctemp = yyFilteredTable->column(*$1);
                                    dssBuilder.currentComponent()->
                                            addFilter(ctemp, *$3);
                                  }else{
                                    dssBuilder.currentComponent()->
                                            addFilter(*$1, *$3);
                                  }
                                  delete $1;
                                  delete $3; 
                                }
        | axisname AND NUMBER RELOP NUMBER 
	                        { if( $4 != EqOpT || $5 != 0 ){
				    selerrstr<<"Bit mask filter must be tested against \"== 0\"."
					     <<selerror("BitMaskTesting");
  				    SELTHROW(DssBuilder::MiscException);
				  }
				  if(yyFilteredTable){
                                    const Column *ctemp = yyFilteredTable->column(*$1);
                                    dssBuilder.currentComponent()->
                                           addFilter(ctemp, ULNG($3));
                                  }else{
                                    dssBuilder.currentComponent()->
                                            addFilter(*$1, ULNG($3));
                                  }
                                  delete $1;
                                }
        | '(' axisname AND NUMBER ')' RELOP NUMBER 
	                        { if( $6 != EqOpT || $7 != 0 ){
				    selerrstr<<"Bit mask filter must be tested against \" == 0\"."
					     <<selerror("MaskSpecOffset");
  				    SELTHROW(DssBuilder::MiscException);
				  }
				  if(yyFilteredTable){
                                    const Column *ctemp = yyFilteredTable->column(*$2);
                                    dssBuilder.currentComponent()->
                                           addFilter(ctemp, ULNG($4));
                                  }else{
                                    dssBuilder.currentComponent()->
                                            addFilter(*$2, ULNG($4));
                                  }
                                  delete $2;
                                }
        | axisname IN BLOCKFILTER1 '(' BLOCK ')'
                                { if(yyFilteredTable){
                                    const Column *ctemp = yyFilteredTable->column(*$1);
                                    dssBuilder.currentComponent()->
                                           addFilter(ctemp, *$5);
                                  }else{
                                    dssBuilder.currentComponent()->
                                            addFilter(*$1, *$5);
                                  }
                                  delete $1;
                                  delete $5;
                                }
        | BLOCKFILTER1 '(' BLOCK ',' axisname ')'
	                        { if(yyFilteredTable){
                                    const Column *ctemp = yyFilteredTable->column(*$5);
                                    dssBuilder.currentComponent()->
                                            addFilter(ctemp, *$3);
                                  }else{
                                    dssBuilder.currentComponent()->
                                            addFilter(*$5, *$3);
                                  }
                                  delete $3;
                                  delete $5;
                                }
        | '(' axisname ',' axisname ')' IN BLOCKFILTER2 '(' BLOCK ')'   
                                { DSS::DSfilterComp fcomp(2);
                                  fcomp[0] = *$2;
                                  fcomp[1] = *$4;
                                  dssBuilder.currentComponent()->
                                           addFilter(dssBuilder.Axis2dName(fcomp), *$9, fcomp);
                                  delete $2;
                                  delete $4;
                                }
        | BLOCKFILTER2 '(' BLOCK ',' axisname ',' axisname ')'
                                { DSS::DSfilterComp fcomp(2);
                                  fcomp[0] = *$5;
                                  fcomp[1] = *$7;
                                  dssBuilder.currentComponent()->
                                           addFilter(dssBuilder.Axis2dName(fcomp), *$3, fcomp);
                                  delete $5;
                                  delete $7;
                                }
        | '(' axisname ',' axisname ')' IN BLOCKFILTER2 '(' BLOCK ',' arith ',' arith ')'   
	                        { if( $11 != 0 || $13 != 0 ){
				    selerrstr<<"Mask filter must have zero offset in the data subspace."
					     <<selerror("MaskSpecOffset");
  				    SELTHROW(DssBuilder::MiscException);
				  }
				  DSS::DSfilterComp fcomp(2);
                                  fcomp[0] = *$2;
                                  fcomp[1] = *$4;
                                  dssBuilder.currentComponent()->
                                           addFilter(dssBuilder.Axis2dName(fcomp), *$9, fcomp);
                                  delete $2;
                                  delete $4;
                                }
        | BLOCKFILTER2 '(' BLOCK ',' arith ',' arith ',' axisname ',' axisname ')'   
	                        { if( $5 != 0 || $7 != 0 ){
				    selerrstr<<"Mask filter must have zero offset in the data subspace."
					     <<selerror("MaskSpecOffset");
  				    SELTHROW(DssBuilder::MiscException);
				  }
				  DSS::DSfilterComp fcomp(2);
                                  fcomp[0] = *$9;
                                  fcomp[1] = *$11;
                                  dssBuilder.currentComponent()->
                                           addFilter(dssBuilder.Axis2dName(fcomp), *$3, fcomp);
                                  delete $9;
                                  delete $11;
                                }
	| axisname RELOP arith
	                        { DRangeList rlist;
				  switch ($2) {
				  case LtOpT:{
				    DRange range(-DBL_MAX, $3, Unbounded, Open);
				    rlist.push_back(range);
				    break; }
				  case LeOpT:{
				    DRange range(-DBL_MAX, $3, Unbounded, Closed);
				    rlist.push_back(range);
				    break; }
				  case GtOpT:{
				    DRange range($3, DBL_MAX, Open, Unbounded);
				    rlist.push_back(range);
				    break; }
				  case GeOpT:{
				    DRange range($3, DBL_MAX, Closed, Unbounded);
				    rlist.push_back(range);
				    break; }
				  case EqOpT:{
				    DRange range($3, $3, Closed, Closed);
				    rlist.push_back(range);
				    break; }
				  case NeOpT:{
				    DRange range1($3, DBL_MAX, Open, Unbounded);
				    DRange range2(-DBL_MAX, $3, Unbounded, Open);
				    rlist.push_back(range1);
				    rlist.push_back(range2);
				    break; }
				  }

				  if(yyFilteredTable){
				    const Column *ctemp = yyFilteredTable->column(*$1);
				    dssBuilder.currentComponent()->
				                   addFilter(ctemp, rlist);
				  }else{
				    dssBuilder.currentComponent()->
				                   addFilter(*$1, rlist);
				  }

                                  delete $1;
				}
	| arith RELOP axisname
	                        { DRangeList rlist;
				  switch ($2) {
				  case GtOpT:{
				    DRange range(-DBL_MAX, $1, Unbounded, Open);
				    rlist.push_back(range);
				    break; }
				  case GeOpT:{
				    DRange range(-DBL_MAX, $1, Unbounded, Closed);
				    rlist.push_back(range);
				    break; }
				  case LtOpT:{
				    DRange range($1, DBL_MAX, Open, Unbounded);
				    rlist.push_back(range);
				    break; }
				  case LeOpT:{
				    DRange range($1, DBL_MAX, Closed, Unbounded);
				    rlist.push_back(range);
				    break; }
				  case EqOpT:{
				    DRange range($1, $1, Closed, Closed);
				    rlist.push_back(range);
				    break; }
                                  case NeOpT:{
				    DRange range1($1, DBL_MAX, Open, Unbounded);
				    DRange range2(-DBL_MAX, $1, Unbounded, Open);
				    rlist.push_back(range1);
				    rlist.push_back(range2);
				    break; }
				  }

				  if(yyFilteredTable){
				    const Column *ctemp = yyFilteredTable->column(*$3);
				    dssBuilder.currentComponent()->
				                   addFilter(ctemp, rlist);
				  }else{
				    dssBuilder.currentComponent()->
				                   addFilter(*$3, rlist);
				  }
                                  delete $3;
				}
        |  regionfilter
                                { GeoRegion r;
                                  r.push_back(
                                        *dssBuilder.currentRegionComponent());
                                  DSS::DSfilterComp fc(2);
                                  fc[0] = dssBuilder.xAxisName();
                                  fc[1] = dssBuilder.yAxisName();
                                  dssBuilder.currentComponent()->
                                          addFilter(dssBuilder.Axis2dName(fc), r, fc);
                                }

regionfilter: NOT '(' regionfilter ')' { dssBuilder.currentRegion()->
                                         include( !dssBuilder.currentRegion()->included() ); }
        |     NOT regionfilter         { dssBuilder.currentRegion()->exclude(); }
        | '(' axisname ',' axisname ')' IN regiontype
                                { dssBuilder.regionAxes(*$2, *$4);
                                  delete $2; delete $4; }
        | RCIRCLE '(' arith ',' arith ',' arith ',' axisname ',' axisname ')'
                  { dssBuilder.addRegion(new FCircle(FPoint($3, $5), $7));
                    dssBuilder.regionAxes(*$9, *$11);
                    delete $9; delete $11;
		  }
        | RRING '(' arith ',' arith ',' arith ',' arith ',' axisname ',' axisname ')'
                  { dssBuilder.addRegion(new FAnnulus(FPoint($3, $5), $7, $9)); 
                    dssBuilder.regionAxes(*$11, *$13);
                    delete $11; delete $13;
		  }
        | RELLIPSE '(' arith ',' arith ',' arith ',' arith ',' arith ',' axisname ',' axisname ')'
                  { dssBuilder.addRegion(new FEllipse(FPoint($3, $5), $7, $9, $11));
                    dssBuilder.regionAxes(*$13, *$15);
                    delete $13; delete $15;
		  }
        | RELLIPSERING '(' arith ',' arith ',' arith ',' arith ',' arith ',' arith ',' arith ',' arith ',' axisname ',' axisname ')'
                  { dssBuilder.addRegion(new FElliptAnnulus(FPoint($3, $5), $7, $9, $11, $13, $15, $17)); 
                    dssBuilder.regionAxes(*$19, *$21);
                    delete $19; delete $21;
		  }
        | RBOX '(' arith ',' arith ',' arith ',' arith ',' arith ',' axisname ',' axisname ')'
                  { dssBuilder.addRegion(new FRotBox(FPoint($3, $5), $7, $9, $11));
                    dssBuilder.regionAxes(*$13, *$15);
                    delete $13; delete $15;
		  }
        | RRECTANGLE '(' arith ',' arith ',' arith ',' arith ',' arith ',' axisname ',' axisname ')'
                  { dssBuilder.addRegion(new FRotRectangle(FPoint($3, $5), $7, $9, $11));
                    dssBuilder.regionAxes(*$13, *$15);
                    delete $13; delete $15;
		  }
        | RDIAMOND '(' arith ',' arith ',' arith ',' arith ',' arith ',' axisname ',' axisname ')'
                  { dssBuilder.addRegion(new FRotRhombus(FPoint($3, $5), $7, $9, $11));
                    dssBuilder.regionAxes(*$13, *$15);
                    delete $13; delete $15;
		  }
        | RSECTOR '(' arith ',' arith ',' arith ',' arith ',' axisname ',' axisname ')'
                  { dssBuilder.addRegion(new FSector(FPoint($3, $5), $7, $9)); 
                    dssBuilder.regionAxes(*$11, *$13);
                    delete $11; delete $13;
		  }
        | RPOLYGON '(' pointlist ',' axisname ',' axisname ')'
                  { dssBuilder.addRegion(new FPolygon($3[0], $3[1], $3[0].size()));
                    delete [] $3; 
                    dssBuilder.regionAxes(*$5, *$7);
                    delete $5; delete $7;
		  }

regiontype: RCIRCLE '(' arith ',' arith ',' arith ')'
          { dssBuilder.addRegion(new FCircle(FPoint($3, $5), $7)); 
	  }
        | RRING '(' arith ',' arith ',' arith ',' arith ')'
          { dssBuilder.addRegion(new FAnnulus(FPoint($3, $5), $7, $9)); 
	  }
        | RELLIPSE '(' arith ',' arith ',' arith ',' arith ',' arith ')'
          { dssBuilder.addRegion(new FEllipse(FPoint($3, $5), $7, $9, $11));
	  }
        | RELLIPSERING '(' arith ',' arith ',' arith ',' arith ',' arith ',' arith ',' arith ',' arith ')'
          { dssBuilder.addRegion(new FElliptAnnulus(FPoint($3, $5), $7, $9, $11, $13, $15, $17)); 
	  }
        | RBOX '(' arith ',' arith ',' arith ',' arith ',' arith ')'
          { dssBuilder.addRegion(new FRotBox(FPoint($3, $5), $7, $9, $11));
	  }
        | RRECTANGLE '(' arith ',' arith ',' arith ',' arith ',' arith ')'
          { dssBuilder.addRegion(new FRotRectangle(FPoint($3, $5), $7, $9, $11));
	  }
        | RDIAMOND '(' arith ',' arith ',' arith ',' arith ',' arith ')'
          { dssBuilder.addRegion(new FRotRhombus(FPoint($3, $5), $7, $9, $11));
	  }
        | RSECTOR '(' arith ',' arith ',' arith ',' arith ')'
          { dssBuilder.addRegion(new FSector(FPoint($3, $5), $7, $9)); 
	  }
        | RPOLYGON '(' pointlist ')'
          { dssBuilder.addRegion(new FPolygon($3[0], $3[1], $3[0].size()));
            delete [] $3; 
	  }

pointlist: pointlist ',' arith ',' arith
                                { $1[0].push_back($3);
                                  $1[1].push_back($5);
                                  $$ = $1; }

        | arith ',' arith       { vector<float> x; x.push_back($1);
                                  vector<float> y; y.push_back($3);
                                  vector<float> *v =
                                        new vector<float>[2];
                                  v[0] = x; v[1] = y;
                                  $$ = v; }


intervallist: intervallist ',' intervalspec
                                { $1->push_back(*$3);
                                  delete $3;
                                  $$ = $1;
                                }
        | intervalspec          { $$ = new DRangeList;
                                  $$->push_back(*$1);
                                  delete $1;
                                }

arith   : arith PLUS arith      { $$ = $1 + $3; }
        | arith MINUS arith     { $$ = $1 - $3; }
        | arith MULT arith      { $$ = $1 * $3; }
        | arith DIV arith       { if ($3 == 0.) {
                                      selerrstr<<"Division by 0 in `/' operator."
                                           <<selerror("ArithmeticException");
                                    SELTHROW(DssBuilder::ArithmeticException);
                                  }
                                  $$ = $1 / $3;
                                }
        | arith MODULO arith    { if ($3 == 0.) {
                                      selerrstr<<"Division by 0 in `modulo' operator."
                                           <<selerror("ArithmeticException");
                                    SELTHROW(DssBuilder::ArithmeticException);
                                  }
                                  $$ = fmod($1, $3);
                                }
        | arith EXPO arith      { double dum;
                                  if ($1 < 0. && modf($3, &dum) != 0.) {
                                      selerrstr<<"Cannot raise "<<$1<<" to the power "<<$3<<" ."
                                           <<selerror("ArithmeticException");
                                    SELTHROW(DssBuilder::ArithmeticException);
                                  }
                                  $$ = pow($1, $3);
                                }
        | arith AND arith       { $$ = DBL(UNT($1) & UNT($3)); }
        | arith OR arith        { $$ = DBL(UNT($1) | UNT($3)); }
        | arith XOR arith       { $$ = DBL(UNT($1) ^ UNT($3)); }
        | arith LSHIFT arith    { $$ = DBL(UNT($1) << UNT($3)); }
        | arith RSHIFT arith    { $$ = DBL(UNT($1) >> UNT($3)); }
        | BITCOMPL arith        { $$ = DBL(~UNT($2)); }
        | AFUNC1 '(' arith ')'  { $$ = (*$1)($3); }
        | AFUNC2 '(' arith ',' arith ')'
                                { $$ = (*$1)($3, $5); }
        | MINUS arith %prec NOT { $$ = -$2; }
        | '(' arith ')'         { $$ = $2; }
        | NUMBER

intervalspec: ':'                       { $$ = new DRange(-DBL_MAX, DBL_MAX,
                                                          Unbounded, Unbounded); }
            | '[' ':' ']'               { $$ = new DRange(-DBL_MAX, DBL_MAX,
                                                          Unbounded, Unbounded); }
            | '(' ':' ']'               { $$ = new DRange(-DBL_MAX, DBL_MAX,
                                                          Unbounded, Unbounded); }
            | '[' ':' ')'               { $$ = new DRange(-DBL_MAX, DBL_MAX,
                                                          Unbounded, Unbounded); }
            | '(' ':' ')'               { $$ = new DRange(-DBL_MAX, DBL_MAX,
                                                          Unbounded, Unbounded); }
            | arith                     { $$ = new DRange($1, $1); }
            | '[' arith ']'             { $$ = new DRange($2, $2); }
            | arith ':'                 { $$ = new DRange($1, DBL_MAX,
                                                          Closed, Unbounded); }
            | '[' arith ':' ']'         { $$ = new DRange($2, DBL_MAX,
                                                          Closed, Unbounded); }
            | '[' arith ':' ')'         { $$ = new DRange($2, DBL_MAX,
                                                          Closed, Unbounded); }
            | '(' arith ':' ']'         { $$ = new DRange($2, DBL_MAX,
                                                          Open, Unbounded); }
            | '(' arith ':' ')'         { $$ = new DRange($2, DBL_MAX,
                                                          Open, Unbounded); }
            | ':' arith                 { $$ = new DRange(-DBL_MAX, $2,
                                                          Unbounded, Closed); }
            | '[' ':' arith ']'         { $$ = new DRange(-DBL_MAX, $3,
                                                          Unbounded, Closed); }
            | '[' ':' arith ')'         { $$ = new DRange(-DBL_MAX, $3,
                                                          Unbounded, Open); }
            | '(' ':' arith ']'         { $$ = new DRange(-DBL_MAX, $3,
                                                          Unbounded, Closed); }
            | '(' ':' arith ')'         { $$ = new DRange(-DBL_MAX, $3,
                                                          Unbounded, Open); }
            | arith ':' arith           { $$ = new DRange($1, $3); }
            | '[' arith ':' arith ']'   { $$ = new DRange($2, $4); }
            | '(' arith ':' arith ']'   { $$ = new DRange($2, $4,
                                                          Open, Closed); }
            | '[' arith ':' arith ')'   { $$ = new DRange($2, $4,
                                                          Closed, Open); }
            | '(' arith ':' arith ')'   { $$ = new DRange($2, $4,
                                                          Open, Open); }
%%

}
