//
//  Implementation of the DSbitFilter class
//
//	AUTHOR:
//	  D. Hinshaw (DH)
//
//	  REFERENCES:
//	    [1] FITS Keyword Conventions in ASC Data Model files, SDS-7.2
// 		J. McDowell
//		http://hea-www.harvard.edu/~jcm/asc/docs/
//
//          [2] ASC Data Model Abstact Design, SDS-3.0
//              J. McDowell
//		http://hea-www.harvard.edu/~jcm/asc/docs/
//	    
//	    [3] Filtering Syntax in the ASC Data Model, SDS-6.0
//	        J. McDowell, A. Fruscione, A. Siemiginowska, B. Joye
//	        http://hea-www.harvard.edu/~jcm/asc/docs/

#include <strstream.h>
#include <cstring>

#include <Dal.h>
#include <Region.h>
#include <dss.h>
#include <errstr.h>


DSbitFilter::DSbitFilter(DScomp *parent,
			 const string &name, unsigned long bitMask) :
  DSfilter(parent, BitMaskT, name), _bitMask(bitMask) {}


DSbitFilter::DSbitFilter(DScomp *parent,
			 const Column *inheritedColumn, 
			 unsigned long bitMask) :
  DSfilter(parent, BitMaskT, inheritedColumn), _bitMask(bitMask) {

    // Check that column is of an integer type
    
    Dal::DataType colType = inheritedColumn->dataType();

    if( colType != Dal::Int8 && 
	colType != Dal::Int16 && 
	colType != Dal::Int32 ){
      errstr << "Column must be an integer for bit mask filters"
	     << Sas::error("DSSbitType");
    }
}


void DSbitFilter::write(unsigned compNumber, unsigned filterNumber, 
			Block *toBlock, bool dummy) const{
  strstream valueName, value, formName;

  value.setf(ios::showbase);
  if( dummy )
    value << ":" << ends;
  else
    value << hex << _bitMask << ends; 
 
  formName << "DSFORM" << filterNumber << ends;
  if( ! toBlock->hasAttribute(formName.str()) )
    *toBlock->addAttribute(formName.str(), comment+"data type") = "X";
    
  if( compNumber==1 )
    valueName << "DSVAL" << filterNumber << ends;   
  else
    valueName << compNumber << "DSVAL" << filterNumber << ends;

  *toBlock->addAttribute(valueName.str(), comment+"value") = value.str();
}


unsigned long DSbitFilter::bitMask() const{
  return _bitMask;
}


string DSbitFilter::selectExpression() {
  strstream expression;
  expression << name() << "&" << _bitMask <<"==0" <<ends;
  return expression.str();
}

 
double DSbitFilter::volume() const{
  errstr << "Cannot calculate volume for a bit mask "
	 << "filter.  Zero volumed returned."
         << Sas::warning("BitVol");
  return 0;
}

bool DSbitFilter::operator==(DSfilter &oFilter) const{

  if( name() != oFilter.name()   ||
      type() != oFilter.type()   ||
      units() != oFilter.units() ||
      comps() != oFilter.comps() )
    return false;
      
  if( bitMask() != oFilter.bitMask() ) return false;

  return true;
}


