//
//  Implementation of the DSblockFilter class
//
//	AUTHOR:
//	  D. Hinshaw (DH)
//
//	  REFERENCES:
//	    [1] FITS Keyword Conventions in ASC Data Model files, SDS-7.2
// 		J. McDowell
//		http://hea-www.harvard.edu/~jcm/asc/docs/
//
//          [2] ASC Data Model Abstact Design, SDS-3.0
//              J. McDowell
//		http://hea-www.harvard.edu/~jcm/asc/docs/
//	    
//	    [3] Filtering Syntax in the ASC Data Model, SDS-6.0
//	        J. McDowell, A. Fruscione, A. Siemiginowska, B. Joye
//	        http://hea-www.harvard.edu/~jcm/asc/docs/

#include <strstream.h>
#include <cstring>

#include <Dal.h>
#include <errstr.h>
#include <Region.h>
#include <ASCregion.h>
#include <dss.h>


void DSblockFilter::_openBlock(const string &blockSpec){
    if( blockSpec.find("[") != string::npos || blockSpec.find("+") != string::npos )
      errstr << "File/block specification '" << blockSpec
	     << "' is not in a format supported by the data subspace."
             << Sas::error("BlockSpec");

    DataSet * blockset = dataSetServer->open(blockSpec, DataSet::Read);

    string::size_type iColon = blockSpec.find(":");
    if( iColon == string::npos )
      _block = blockset->block(0);
    else
      _block = blockset->block(blockSpec);
}


DSblockFilter::DSblockFilter(DScomp *parent,
			     const FilterType ftype, const string &name, 
			     const string &blockSpec,
			     const string &units = "") :
  DSfilter(parent, ftype, name, units) { _openBlock(blockSpec); }


DSblockFilter::DSblockFilter(DScomp *parent,
			     const FilterType ftype,
			     const Column *inheritedColumn,
			     const string &blockSpec) :
  DSfilter(parent, ftype, inheritedColumn) { _openBlock(blockSpec); }


DSblockFilter::DSblockFilter(DScomp *parent,
			     const FilterType ftype,
			     const string &name,
			     const string &blockSpec,
			     DSfilterComp &filtercomp) :
  DSfilter(parent, ftype, name, "", filtercomp) { _openBlock(blockSpec); }


DSblockFilter::DSblockFilter(DSblockFilter &dsblock) :
  DSfilter(dsblock) {
    DataSet * blockset = dataSetServer->open(dsblock._block->parent()->name(), DataSet::Read);
    _block = blockset->block(dsblock._block->name());
}
 
 
string DSblockFilter::blockSpec() const {
  return _block->parent()->name() + ":" +  _block->name();
}

void DSblockFilter::write(unsigned compNumber, unsigned filterNumber, 
			  Block *toBlock, bool dummy) const{
  strstream valueName0;

  string value, ref = _block->name();

  valueName0 << "DSVAL" << filterNumber << ends;

  if( type() == MaskT ) 
    value = "IMAGE";
  else
    value = "TABLE";

  if( !toBlock->hasAttribute(valueName0.str()) )
    *toBlock->addAttribute(valueName0.str(), comment+"value") = value;

  if( type() == MaskT || type() == RegionBlockT ){
    DSfilterComp filterComps = comps();
    strstream typeName;
    typeName << "DSTYP" << filterNumber << ends;
    
    *toBlock->attribute(typeName.str()) = toBlock->attribute(typeName.str())->asString() + 
      "(" + filterComps[0] + "," + filterComps[1] + ")";
  }

  string refValue;
  if( dummy )
    refValue = ":";
  else{
    bool goodRef = false;
    while( !goodRef ){
      refValue = ":" + ref;

      // Need to copy the filter block to the output data set.
      // First check for a name clash

      bool prevRef = false;
      for(unsigned iComp=1; iComp<compNumber; iComp++){
	strstream pRefName;
	if( iComp==1 )
	  pRefName << "DSREF" << filterNumber << ends;
	else
	  pRefName << iComp << "DSREF" << filterNumber << ends;

	if( toBlock->hasAttribute(pRefName.str()) &&
	    toBlock->attribute(pRefName.str())->asString() == refValue ){
	
	  if( parent()->parent()->component(iComp-1)
	      ->filter(*this)->blockSpec() != blockSpec() )
	    errstr << "The filter " << name() 
		   << " references different blocks with the same name. "
		   << "Cannot write data subspace." << Sas::error("DssRefClash");
	
	  prevRef = true;
	}
      }

      if( toBlock->parent()->hasBlock(ref) ){
	if( toBlock->parent()->name()+refValue != blockSpec() && !prevRef ){
	  errstr << "Name clash while writing Data Subspace: Block "
		 << ref << " already exists in output data set.";

	  if( ref.substr(0,6) == "DSSBLK" ){
	    strstream blkNumStr;
	    int blkNum = atoi( ref.substr(7,2).c_str() ) + 1;
	    blkNumStr << "DSSBLK";
	    blkNumStr.width(2);
	    blkNumStr.fill('0');
	    blkNumStr << blkNum << ends;
	    ref = blkNumStr.str();
	  }else{
	    ref = "DSSBLK01";
	  }

	  errstr << "Renaming block to " << ref << "." << Sas::warning("DssBlockClash");
	}else{
	  goodRef = true;
	}
      }else{
	toBlock->parent()->add(_block, ref);
	goodRef = true;
      }
    }
  }

  strstream refName;

  if( compNumber==1 )
    refName << "DSREF" << filterNumber << ends;
  else
    refName << compNumber << "DSREF" << filterNumber << ends;

  *toBlock->addAttribute(refName.str(), comment+"reference") = refValue;

  return;
}


bool DSblockFilter::operator==(DSfilter &oFilter) const{

  if( name() != oFilter.name()   ||
      type() != oFilter.type()   ||
      units() != oFilter.units() ||
      comps() != oFilter.comps() )
    return false;
      
  if( blockSpec() != oFilter.blockSpec() ) return false;

  return true;
}


DSblockFilter::~DSblockFilter() {
  dataSetServer->close(_block->parent());
}
