//
//  Implementation of the DScomp class
//
//	AUTHOR:
//	  D. Hinshaw (DH)
//
//	  REFERENCES:
//	    [1] FITS Keyword Conventions in ASC Data Model files, SDS-7.2
// 		J. McDowell
//		http://hea-www.harvard.edu/~jcm/asc/docs/
//
//          [2] ASC Data Model Abstact Design, SDS-3.0
//              J. McDowell
//		http://hea-www.harvard.edu/~jcm/asc/docs/
//	    
//	    [3] Filtering Syntax in the ASC Data Model, SDS-6.0
//	        J. McDowell, A. Fruscione, A. Siemiginowska, B. Joye
//	        http://hea-www.harvard.edu/~jcm/asc/docs/


#include <strstream.h>
#include <cstring>
#include <math.h>

#include <Dal.h>
#include <Region.h>
#include <dss.h>
#include <errstr.h>


DScomp::DScomp(DataSubspace *parent) : 
  DSchild<DataSubspace>(parent), _filtvec(0) {}


DScomp::DScomp(const DScomp &dscomp) : DSchild<DataSubspace>(dscomp.parent()) {

  for(unsigned ifilter=0; ifilter<dscomp.filters(); ifilter++)
    addFilter( *dscomp.filter(ifilter) );

  return;
}


DScomp &DScomp::operator=(const DScomp &dscomp){
  vector<DSfilter *>::iterator ifilt;
  for(ifilt=_filtvec.begin(); ifilt!=_filtvec.end(); ifilt++)
    delete *ifilt;
  _filtvec.clear();

  for(unsigned ifilter=0; ifilter<dscomp.filters(); ifilter++)
    addFilter( *dscomp.filter(ifilter) );

  return *this;
}


// Add a copy of a filter
DSfilter * DScomp::addFilter(DSfilter &filter){
  DSfilter * newFilter;

  switch( filter.type() ){
  case DSS::RangeT:
    if( dynamic_cast<DSintRange<long> *>(&filter) )
      newFilter = new DSintRange<long>( dynamic_cast<DSintRange<long> &>(filter) );
    else
      newFilter = new DSrealRange<double>( dynamic_cast<DSrealRange<double> &>(filter) );
    break;
  case DSS::RegionT:
    newFilter = new DSregionFilter( dynamic_cast<DSregionFilter &>(filter) );
    break;
  case DSS::BitMaskT:
    newFilter = new DSbitFilter( dynamic_cast<DSbitFilter &>(filter) ); 
    break;
  case DSS::RangeBlockT:
    newFilter = new DSrangeBlock( dynamic_cast<DSrangeBlock &>(filter) );
    break;
  case DSS::GtiT:
    newFilter = new DSgtiBlock( dynamic_cast<DSgtiBlock &>(filter) );
    break;
  case DSS::MaskT:
    newFilter = new DSmaskBlock( dynamic_cast<DSmaskBlock &>(filter) );
    break;
  case DSS::RegionBlockT:
    newFilter = new DSregionBlock( dynamic_cast<DSregionBlock &>(filter) );
    break;
  }

  newFilter->parent(this);
  _nameCheck(*newFilter);
  _filtvec.push_back(newFilter);

  return newFilter;
}

// Add range filters

DSfilter * DScomp::addFilter(const string &name,
		       const list<Range<long> > ranges,
		       const string &units = ""){

  DSfilter * filter = new DSintRange<long>(this, name, ranges, units);
  _nameCheck(*filter);
  _filtvec.push_back(filter);

  return filter;
}


DSfilter * DScomp::addFilter(const string &name,
		       const list<Range<double> > ranges,
		       const string &units = ""){

  DSfilter * filter = new DSrealRange<double>(this, name, ranges, units);
  _nameCheck(*filter);
  _filtvec.push_back(filter);

  return filter;
}


DSfilter *DScomp::addFilter(const Column *inheritedColumn,
		      const list<Range<long> > ranges){
  DSfilter * filter;

  Dal::DataType colType = inheritedColumn->dataType();
  if( colType == Dal::Real32 ||
      colType == Dal::Real64 ){
    list<Range<double> > dranges;
    list<Range<long> >::const_iterator irange;

    for(irange=ranges.begin(); irange!=ranges.end(); irange++){
      Range<double> drange(*irange);
      dranges.push_back(drange);
    }
    filter = new DSrealRange<double>(this, inheritedColumn, dranges);

  }else
    filter = new DSintRange<long>(this, inheritedColumn, ranges);

  _nameCheck(*filter);
  _filtvec.push_back(filter);

  return filter;
}


DSfilter *DScomp::addFilter(const Column *inheritedColumn,
		      const list<Range<double> > ranges){
  DSfilter * filter;

  Dal::DataType colType = inheritedColumn->dataType();
  if( colType == Dal::Int8  ||
      colType == Dal::Int16 || 
      colType == Dal::Int32 ){
    list<Range<double> >::const_iterator irange;

    // Check to see if the real ranges are really integer values

    for(irange=ranges.begin(); irange!=ranges.end(); irange++){
      RangeBound lowerB, upperB;
      double lower, upper;
      irange->values(lower, upper, lowerB, upperB);
      if( floor(lower) != lower || floor(upper) != upper )
	errstr << "Real-valued ranges used for integer-valued columns. "
	       << "Real values will be truncated to integers."
	       << Sas::warning("DSStypeTruncate");
    }

    list<Range<long> > intRanges;
 
    for(irange=ranges.begin(); irange!=ranges.end(); irange++){
      Range<long> intRange(*irange);
      intRanges.push_back(intRange);
    }
    filter = new DSintRange<long>(this, inheritedColumn, intRanges);
    
  }else
    filter = new DSrealRange<double>(this, inheritedColumn, ranges);

  _nameCheck(*filter);
  _filtvec.push_back(filter);

  return filter;
}

// Add bitmask filters

DSfilter *DScomp::addFilter(const string &name,
		      unsigned long bitMask){

  DSfilter * filter = new DSbitFilter(this, name, bitMask);
  _nameCheck(*filter);
  _filtvec.push_back(filter);

  return filter;
}


DSfilter *DScomp::addFilter(const Column *inheritedColumn,
		      unsigned long bitMask){

  DSfilter * filter = new DSbitFilter(this, inheritedColumn, bitMask);
  _nameCheck(*filter);
  _filtvec.push_back(filter);

  return filter;
}

// Add block filters

DSfilter *DScomp::addFilter(const string &name,
		      const string &blockSpec,
		      const string &units = ""){

  DataSet * filtset = dataSetServer->open(blockSpec, DataSet::Read);
  Block * filterBlock;

  string::size_type iColon = blockSpec.find(":");
  if( iColon == string::npos )
    filterBlock = filtset->block(0);
  else
    filterBlock = filtset->block(blockSpec);

  // Query the information in the block to determine what
  // kind of filter this is.

  DSfilter * filter;

  if( filterBlock->type() == Block::ArrayT ){

    filter = new DSmaskBlock(this, name, blockSpec);

  }else{
    Table * filterTable = dynamic_cast<Table *>(filterBlock);

    if( filterBlock->hasAttribute("HDUCLASS") &&
	filterBlock->hasAttribute("HDUCLAS1") ){

      if( filterBlock->attribute("HDUCLASS")->asString() == "ASC" &&
	  filterBlock->attribute("HDUCLAS1")->asString() == "REGION" )
	filter = new DSregionBlock(this, name, blockSpec);
      else if( filterBlock->attribute("HDUCLASS")->asString() == "OGIP" &&
	       filterBlock->attribute("HDUCLAS1")->asString() == "GTI" )
	filter = new DSgtiBlock(this, name, blockSpec, units);

      // So much for the easy way of determining the block type, now try
      // the hard way.  If it has START and STOP columns, it is probably
      // a GTI.

    }else if( filterTable->hasColumn("START") && filterTable->hasColumn("STOP") )
      filter = new DSgtiBlock(this, name, blockSpec, units);

    else{
      // Check if the block could be a range table.

      for(unsigned i=1;;i++){
	strstream mtype;
	mtype << "MTYPE" << i << ends;

	if( filterBlock->hasAttribute(mtype.str()) ){
	  if( filterBlock->attribute(mtype.str())->asString() == name ){
	    filter = new DSrangeBlock(this, name, blockSpec, units);
	    break;
	  }
	}else
	  errstr << "Cannot recognize block " << filterBlock->name()
		 << " as a valid filter type." << Sas::error("DSSblockType");
      }
    }
  }

  _nameCheck(*filter);
  _filtvec.push_back(filter);

  dataSetServer->close(filtset);
  return filter;
}


DSfilter *DScomp::addFilter(const Column *inheritedColumn,
			    const string &blockSpec){

  DataSet * filtset = dataSetServer->open(blockSpec, DataSet::Read);
  Block * filterBlock;

  string::size_type iColon = blockSpec.find(":");
  if( iColon == string::npos )
    filterBlock = filtset->block(0);
  else
    filterBlock = filtset->block(blockSpec);

  // Query the information in the block to determine what
  // kind of filter this is.
  DSfilter * filter;

  if( filterBlock->hasAttribute("HDUCLASS") &&
      filterBlock->hasAttribute("HDUCLAS1") &&
      filterBlock->attribute("HDUCLASS")->asString() == "OGIP" &&
      filterBlock->attribute("HDUCLAS1")->asString() == "GTI" ) 
      
    filter = new DSgtiBlock(this, inheritedColumn, blockSpec);

  else{
      // So much for the easy way of determining the block type, now try
      // the hard way.  If it has START and STOP columns, it is probably
      // a GTI.

      Table * filterTable = dynamic_cast<Table *>(filterBlock);
      if( filterTable->hasColumn("START") && filterTable->hasColumn("STOP") )
	filter = new DSgtiBlock(this, inheritedColumn, blockSpec);
      else{

	// Check if the block could be a range table.

	for(unsigned i=1;;i++){
	  strstream mtype;
	  mtype << "MTYPE" << i << ends;

	  if( filterBlock->hasAttribute(mtype.str()) ){
	    if( filterBlock->attribute(mtype.str())->asString() == inheritedColumn->name() ){
	      filter = new DSrangeBlock(this, inheritedColumn, blockSpec);
	      break;
	    }
	  }else
            errstr << "Cannot recognize block " << filterBlock->name()
		   << " as a valid filter type." << Sas::error("DSSblockType");
	}
      }
	
  }

  _nameCheck(*filter);
  _filtvec.push_back(filter);

  dataSetServer->close(filtset);
  return filter;
}


DSfilter *DScomp::addFilter(const string &name,
		      const string &blockSpec,
		      DSfilterComp &filtercomp){

  DataSet * filtset = dataSetServer->open(blockSpec, DataSet::Read);
  Block * filterBlock;

  string::size_type iColon = blockSpec.find(":");
  if( iColon == string::npos )
    filterBlock = filtset->block(0);
  else
    filterBlock = filtset->block(blockSpec);

  // Query the information in the block to determine what
  // kind of filter this is.
  DSfilter * filter;

  if( filterBlock->type() == Block::ArrayT ){
    filter = new DSmaskBlock(this, name, blockSpec, filtercomp);
  }else{
    filter = new DSregionBlock(this, name, blockSpec, filtercomp);
  }

  _nameCheck(*filter);
  _filtvec.push_back(filter);

  dataSetServer->close(filtset);
  return filter;
}

// Add region filters

DSfilter *DScomp::addFilter(const string &name,
		      GeoRegion region,
		      DSfilterComp &filtercomp){
  
  DSfilter * filter = new DSregionFilter(this, name, region, filtercomp);
  _nameCheck(*filter);
  _filtvec.push_back(filter);

  return filter;
}

// Add coordinate filters

DSfilter *DScomp::addFilter(const string &name,
	      GeoRegion region){

  DSfilter * filter = new DSregionFilter(this, name, region);
  _nameCheck(*filter);
  _filtvec.push_back(filter);

  return filter;
}

// Other methods

void DScomp::deleteFilter(unsigned filterNumber){
  delete *( _filtvec.begin()+filterNumber );
  _filtvec.erase( _filtvec.begin()+filterNumber );
}


DSfilter *DScomp::filter(unsigned filterNumber) const{
  return _filtvec[filterNumber];
}


 DSfilter *DScomp::filter(const string &filterName, 
			  const FilterType &filterType) const{
   vector<DSfilter *const>::iterator ifilt;
 
   for(ifilt=_filtvec.begin(); ifilt!=_filtvec.end(); ifilt++){
     if( (*ifilt)->name() == filterName && (*ifilt)->type() == filterType )
       return *ifilt;
   }
 
   errstr << "Filter " << filterName 
 	 << " not found in data subspace component."
 	 << Sas::error("NoFilter");
 
   return 0;
 }
 
 
 bool DScomp::hasFilter(const string &filterName, 
			const FilterType &filterType) const{
   vector<DSfilter *const>::iterator ifilt;
 
   for(ifilt=_filtvec.begin(); ifilt!=_filtvec.end(); ifilt++){
     if( (*ifilt)->name() == filterName && (*ifilt)->type() == filterType )
       return true;
   }
 
   return false;
 }


DSfilter *DScomp::filter(const DSfilter &filter) const{
  vector<DSfilter *const>::iterator ifilt;

  for(ifilt=_filtvec.begin(); ifilt!=_filtvec.end(); ifilt++){
    if( (*ifilt)->like(filter) )
      return *ifilt;
  }

  errstr << "Filter " << filter.name()
	 << " not found in data subspace component."
	 << Sas::error("NoFilter");

  return 0;
}


bool DScomp::hasFilter(const DSfilter &filter) const{
  vector<DSfilter *const>::iterator ifilt;

  for(ifilt=_filtvec.begin(); ifilt!=_filtvec.end(); ifilt++){
    if( (*ifilt)->like(filter) )
      return true;
  }

  return false;
}


unsigned DScomp::filters() const{
  return _filtvec.size();
}


string DScomp::selectExpression(const DSaxes &axes=DataSubspace::noAxes) const{
  vector<DSfilter *const>::iterator ifilt;
  DSaxes::const_iterator iaxis;
  strstream expression;
  
  bool firstfilt = true;
  for(ifilt=_filtvec.begin(); ifilt!=_filtvec.end(); ifilt++){
    if( axes!=DataSubspace::noAxes ){

      for(iaxis=axes.begin(); iaxis!=axes.end(); iaxis++){
	if( *iaxis==(*ifilt)->name() ){
	  if(firstfilt) firstfilt=false;
	  else expression << " && ";
	  expression << (*ifilt)->selectExpression();
	  break;
	}
      }

      // If both of the components of a 2D axis are contained in
      // the list of axes, then project it too.

      if( (*ifilt)->type()==RegionT || (*ifilt)->type()==RegionBlockT ){
	DSfilterComp comps = (*ifilt)->comps();
	unsigned hasComp = 0;
	for(iaxis=axes.begin(); iaxis!=axes.end(); iaxis++){
	  if( *iaxis==comps[0] || *iaxis==comps[1] ) hasComp++;
	}
	if( hasComp==2 ){
	  if(firstfilt) firstfilt=false;
	  else expression << " && ";
	  expression << (*ifilt)->selectExpression();
	}
      }

    }else{
      if( ifilt!=_filtvec.begin() ) expression << " && ";
      expression << (*ifilt)->selectExpression();
    }
  }

  if( expression.str()==NULL ) return "true";

  expression << ends;
  return expression.str();
}


Range<double> DScomp::extent(const string &axis) const{
    vector<DSfilter *const>::iterator ifilt;
    list<Range<double> > rlist;
    
    for(ifilt=_filtvec.begin(); ifilt!=_filtvec.end(); ifilt++){
      if((*ifilt)->name() == axis) rlist.push_back( (*ifilt)->extent(axis) );
      if( (*ifilt)->type()==RegionT || (*ifilt)->type()==RegionBlockT ){
	DSfilterComp lcomps = (*ifilt)->comps();
	if( lcomps[0]==axis || lcomps[1]==axis ) rlist.push_back( (*ifilt)->extent(axis) );
      }
    }

    if( rlist.size()==0 ){
      Range<double> r(0,0,Unbounded,Unbounded);
      return r;
    }

    return maxRange(rlist);
}


double DScomp::volume(const DSaxes &axes) const{
  vector<DSfilter *const>::iterator ifilt;
  DSaxes::const_iterator iaxis;
  double volume=1;

  for(iaxis=axes.begin(); iaxis!=axes.end(); iaxis++){
    double avol = 0;
    for(ifilt=_filtvec.begin(); ifilt!=_filtvec.end(); ifilt++){
      if( *iaxis==(*ifilt)->name() ){
	avol += (*ifilt)->volume();
	break;
      }
    }
    volume *= avol;
  }
  
  return volume;
}


list<Range<double> > DScomp::ranges(const string &axis) const{
  vector<DSfilter *const>::const_iterator ifilt;
  list<Range<double> > r;

  for(ifilt=_filtvec.begin(); ifilt!=_filtvec.end(); ifilt++){
    if( (*ifilt)->name()==axis ){
      list<Range<double> > rtemp = (*ifilt)->ranges();
      r.merge(rtemp);
    }
  }

  return r;
}


//void DScomp::write(unsigned compNumber, Block *toBlock) const{
//  vector<DSfilter *>::iterator ifilt;
//  unsigned i;
//  
//  for(ifilt=_filtvec.begin(), i=0; ifilt!=_filtvec.end(); ifilt++, i++)
//    (*ifilt)->write(compNumber, i, toBlock);
//}


DScomp::~DScomp(){
  vector<DSfilter *>::iterator ifilt;

  for(ifilt=_filtvec.begin(); ifilt!=_filtvec.end(); ifilt++)
    delete *ifilt;

  return;
}

void DScomp::_nameCheck(DSfilter &filter){
  vector<DSfilter *>::iterator ifilt;
  unsigned ident=0;

  for(ifilt=_filtvec.begin(); ifilt!=_filtvec.end(); ifilt++)
    if( filter._name==(*ifilt)->_name     &&
	filter._ftype==(*ifilt)->_ftype   &&
	filter._units==(*ifilt)->_units   &&
	filter.comps()==(*ifilt)->comps() &&
	ident < (*ifilt)->_identifier     )
      ident = (*ifilt)->_identifier;

  filter._identifier = ident+1;
  return;
}


// Perform union of two subspace components.
DScomp &DScomp::operator*=(const DScomp &comp){
  for(unsigned i=0; i<comp.filters(); i++)
      addFilter(*comp.filter(i));
  return *this;
}


void DScomp::compressRegions() {

  // For now we only handle the simple case where each GeoRegion has one Component

  bool compressed = true;
  while( compressed ){
    compressed = false;
    for(unsigned ifilt1=0; ifilt1<filters(); ifilt1++){
      DSfilter *filter1 = filter(ifilt1);
      if( filter1->type() != RegionT ||
	  filter1->region().size() != 1 )continue;
      DSfilterComp comp1 = filter1->comps();
      for(unsigned ifilt2=ifilt1+1; ifilt2<filters(); ifilt2++){
	if( filter(ifilt2)->type() != RegionT ||
	    filter(ifilt2)->comps() != comp1 ||
	    filter(ifilt2)->region().size() != 1  )continue;
	compressed = true;

	GeoRegion region2 = filter(ifilt2)->region();
	GeoRegion region1 = filter1->region();
	for(unsigned ishape=0; ishape<region2[0].size(); ishape++)
	  region1[0].push_back(region2[0][ishape]);
	addFilter(filter1->name(), region1, comp1);

	deleteFilter(ifilt2);
	deleteFilter(ifilt1);
	break;
      }
      if(compressed) break;
    }
  }

  return;
}


DScomp operator*(const DScomp &dscomp1, const DScomp &dscomp2){
  DScomp dscompSum = dscomp1;
  return dscompSum *= dscomp2;
}
