//
//  IMplementation of the DSfilter class
//
//	AUTHOR:
//	  D. Hinshaw (DH)
//
//	  REFERENCES:
//	    [1] FITS Keyword Conventions in ASC Data Model files, SDS-7.2
// 		J. McDowell
//		http://hea-www.harvard.edu/~jcm/asc/docs/
//
//          [2] ASC Data Model Abstact Design, SDS-3.0
//              J. McDowell
//		http://hea-www.harvard.edu/~jcm/asc/docs/
//	    
//	    [3] Filtering Syntax in the ASC Data Model, SDS-6.0
//	        J. McDowell, A. Fruscione, A. Siemiginowska, B. Joye
//	        http://hea-www.harvard.edu/~jcm/asc/docs/


#include <strstream.h>
#include <cstring>

#include <Dal.h>
#include <Region.h>
#include <dss.h>
#include <errstr.h>


const DSS::DSfilterComp DSS::noComps(0);

const string DSfilter::comment("data subspace descriptor: ");


DSfilter::DSfilter(DScomp *parent,
		   const FilterType ftype, const string &name, 
		   const string &units = "",
		   const DSfilterComp &filtercomp = noComps) :
  DSchild<DScomp>(parent), 
  _ftype(ftype), _name(name), _units(units), _comps(filtercomp), _identifier(0) {}


DSfilter::DSfilter(DScomp *parent,
		   const FilterType ftype, const Column *inheritedColumn) : 
  DSchild<DScomp>(parent), _ftype(ftype), _identifier(0) {

  _name = inheritedColumn->name();
  _units = inheritedColumn->units();
}


DSfilter::FilterType DSfilter::type() const{
  return _ftype;
}


string DSfilter::name() const{
  return _name;
}


string DSfilter::units() const{
  return _units;
}


DSS::DSfilterComp DSfilter::comps() const{
  DSS::DSfilterComp lcomps;

  if( _comps!=noComps ){
    lcomps = _comps;
  }else{
    if( _name=="POS" ){
      lcomps.push_back("X");
      lcomps.push_back("Y");

    }else if( _name=="EQPOS" ){
      lcomps.push_back("RA");
      lcomps.push_back("DEC");

    }else if( _name=="GALPOS" ){
      lcomps.push_back("GLON");
      lcomps.push_back("GLAT");

    }else{
      lcomps.push_back(_name+"X");
      lcomps.push_back(_name+"Y");
    }
  }

  return lcomps;
}


bool DSfilter::like(const DSfilter &ofilter) const{
  if( ofilter._name==_name     &&
      ofilter._ftype==_ftype   &&
      ofilter._units==_units   &&
      ofilter.comps()==comps() && 
      ofilter._identifier==_identifier)
    return true;

  return false;
}



list<Range<double> > DSfilter::ranges() const{
  errstr << "No range information available for this filter type." 
	 << Sas::error("NoRange");

  list<Range<double> > dummy;
  return dummy;
}


Range<double> DSfilter::extent(const string &axis) const{
  if( axis!=_name )
    errstr << "Attempt to extract extent information from an unsupported filter"
	   << Sas::error("DSextentType");

  Range<double> r(0,0,Unbounded,Unbounded);
  return r;
}


unsigned long DSfilter::bitMask() const{
  errstr << "No bit mask information available for this filter type." 
	 << Sas::error("NoMask");

  unsigned long dummy = 0;
  return dummy;
}


GeoRegion DSfilter::region() const{
  errstr << "No region information available for this filter type." 
	 << Sas::error("NoRegion");

  GeoRegion dummy;
  return dummy;
}


string DSfilter::blockSpec() const{
  errstr << "No block information available for this filter type." 
	 << Sas::error("NoBlock");

  return " ";
}
