//
//  Implementation of the DSgtiBLock class
//
//	AUTHOR:
//	  D. Hinshaw (DH)
//
//	  REFERENCES:
//	    [1] FITS Keyword Conventions in ASC Data Model files, SDS-7.2
// 		J. McDowell
//		http://hea-www.harvard.edu/~jcm/asc/docs/
//
//          [2] ASC Data Model Abstact Design, SDS-3.0
//              J. McDowell
//		http://hea-www.harvard.edu/~jcm/asc/docs/
//	    
//	    [3] Filtering Syntax in the ASC Data Model, SDS-6.0
//	        J. McDowell, A. Fruscione, A. Siemiginowska, B. Joye
//	        http://hea-www.harvard.edu/~jcm/asc/docs/

#include <strstream.h>
#include <cstring>

#include <Dal.h>
#include <Region.h>
#include <dss.h>
#include <errstr.h>


DSgtiBlock::DSgtiBlock(DSgtiBlock &dsblock) : DSblockFilter(dsblock) {}


DSgtiBlock::DSgtiBlock(DScomp *parent,
		       const string &name, const string &blockSpec,
		       const string &units = "") :
  DSblockFilter(parent, GtiT, name, blockSpec, units){

    // Check that this block contains a valid gti table

    if( _block->type() == Block::TableT  &&
	_block->hasAttribute("HDUCLASS") &&
	_block->hasAttribute("HDUCLAS1") &&
	_block->attribute("HDUCLASS")->asString() == "OGIP" &&
	_block->attribute("HDUCLAS1")->asString() == "GTI" )
      return;
    else{
      Table * dsTable = dynamic_cast<Table *>(_block);
      if( dsTable->hasColumn("START") && dsTable->hasColumn("STOP") )
	return;
    }

    errstr << "Block type is not compatible with the gti filter type"
	   << Sas::error("BlockTypeGti");
}


DSgtiBlock::DSgtiBlock(DScomp *parent,
		       const Column *inheritedColumn, const string &blockSpec) :
  DSblockFilter(parent, GtiT, inheritedColumn, blockSpec){

    // Check that this block contains a valid gti table

    if( _block->type() == Block::TableT  &&
	_block->hasAttribute("HDUCLASS") &&
	_block->hasAttribute("HDUCLAS1") &&
	_block->attribute("HDUCLASS")->asString() == "OGIP" &&
	_block->attribute("HDUCLAS1")->asString() == "GTI" )
      return;
    else{
      Table * dsTable = dynamic_cast<Table *>(_block);
      if( dsTable->hasColumn("START") && dsTable->hasColumn("STOP") )
	return;
    }

    errstr << "Block type is not compatible with the gti filter type"
	   << Sas::error("BlockTypeGti");
}


string DSgtiBlock::selectExpression() {
  strstream expression;
  expression << "gti(" << _block->parent()->name() << ":" << _block->name() << "," 
	     << _name << ")" << ends;
  return expression.str();
}


Range<double> DSgtiBlock::extent(const string &axis) const{
  if( axis!=_name )
    errstr << "Attempt to extract extent information from the wrong filter"
	   << Sas::error("DSextentGTI");

  Table * table = dynamic_cast<Table *>(_block);

  ColumnData *lowerData = table->column("START")->data();
  ColumnData *upperData = table->column("STOP")->data();

  long unsigned nrows = table->rows();

  real64 *lowerVal = lowerData->real64Data();
  real64 *upperVal = upperData->real64Data();

  real64 min=lowerVal[0], max=upperVal[0];

  for(long unsigned irow=1; irow<nrows; irow++){
    if(lowerVal[irow]<min) min = lowerVal[irow];
    if(upperVal[irow]>max) max = upperVal[irow];
  }

  delete lowerVal;
  delete upperVal;

  Range<double> r(min,max);
  return r;
}

  
double DSgtiBlock::volume() const{
  double vol=0;

  Table * table = dynamic_cast<Table *>(_block);

  ColumnData *lowerData = table->column("START")->data();
  ColumnData *upperData = table->column("STOP")->data();

  long unsigned nrows = table->rows();

  real64 *lowerVal = lowerData->real64Data();
  real64 *upperVal = upperData->real64Data();

  for(long unsigned irow=0; irow<nrows; irow++){
    vol += upperVal[irow]-lowerVal[irow]+1;
  }

  delete lowerVal;
  delete upperVal;

  return vol;  
}
