//
//  Implementation of the DSmaskBLock class
//
//	AUTHOR:
//	  D. Hinshaw (DH)
//
//	  REFERENCES:
//	    [1] FITS Keyword Conventions in ASC Data Model files, SDS-7.2
// 		J. McDowell
//		http://hea-www.harvard.edu/~jcm/asc/docs/
//
//          [2] ASC Data Model Abstact Design, SDS-3.0
//              J. McDowell
//		http://hea-www.harvard.edu/~jcm/asc/docs/
//	    
//	    [3] Filtering Syntax in the ASC Data Model, SDS-6.0
//	        J. McDowell, A. Fruscione, A. Siemiginowska, B. Joye
//	        http://hea-www.harvard.edu/~jcm/asc/docs/

#include <strstream.h>
#include <cstring>

#include <Dal.h>
#include <Region.h>
#include <dss.h>
#include <errstr.h>


DSmaskBlock::DSmaskBlock(DSmaskBlock &dsblock) : DSblockFilter(dsblock) {}


DSmaskBlock::DSmaskBlock(DScomp *parent,
			 const string &name, const string &blockSpec,
			 DSfilterComp &filtercomp) :
  DSblockFilter(parent, MaskT, name, blockSpec, filtercomp) {

    if( _block->type() == Block::ArrayT )
      return;
  
    errstr << "Block type is not compatible with the mask filter type"
	   << Sas::error("BlockTypeMask");
}


DSmaskBlock::DSmaskBlock(DScomp *parent,
			 const string &name, const string &blockSpec) :
  DSblockFilter(parent, MaskT, name, blockSpec) {

    // Check that the block type is compatible

    if( _block->type() == Block::ArrayT )
      return;

    errstr << "Block type is not compatible with the mask filter type"
	   << Sas::error("BlockTypeMask");
}


string DSmaskBlock::selectExpression() {
  strstream expression;
  DSfilterComp lcomps = comps();

  expression << "mask(" << _block->parent()->name() << ":" << _block->name() << ",0,0," 
	     <<  lcomps[0] << "," << lcomps[1] << ")" << ends;

  return expression.str();
}


double DSmaskBlock::volume() const{
  double vol = 0.;

  Array * array = dynamic_cast<Array *>(_block);

  ArrayData * aData = array->data();
  vector<unsigned long> aSize = array->size();

  #define CODE(type)                                    \
  {                                                     \
    type *values = aData->type##Data();                 \
                                                        \
    for(long unsigned i=0; i<aSize[0]*aSize[1]; i++){   \
      if( values[i]!=0 ) vol += 1;                      \
    }                                                   \
    break;                                              \
  }                                                     \

  switch( array->dataType()){

  case Dal::Int32  : CODE(int32)
		      
  case Dal::Int16  : CODE(int16)

  case Dal::Int8   : CODE(int8)

  case Dal::Real32 : CODE(real32)

  case Dal::Real64 : CODE(real64)

  default :
    errstr << "Mask array is not of a supported data type." << Sas::error("MaskType");
  }

  #undef CODE

  delete aData;

  if( array->hasAttribute("CRPIX1") && array->hasAttribute("CRPIX2") ){
    vol *= array->attribute("CRPIX1")->asReal() *
      array->attribute("CRPIX2")->asReal();
  } 
  
  return vol;
}
