//
//  IMplementation of the DSrangeFilter, DSintRange and DSrealRange filters
//
//	AUTHOR:
//	  D. Hinshaw (DH)
//
//	  REFERENCES:
//	    [1] FITS Keyword Conventions in ASC Data Model files, SDS-7.2
// 		J. McDowell
//		http://hea-www.harvard.edu/~jcm/asc/docs/
//
//          [2] ASC Data Model Abstact Design, SDS-3.0
//              J. McDowell
//		http://hea-www.harvard.edu/~jcm/asc/docs/
//	    
//	    [3] Filtering Syntax in the ASC Data Model, SDS-6.0
//	        J. McDowell, A. Fruscione, A. Siemiginowska, B. Joye
//	        http://hea-www.harvard.edu/~jcm/asc/docs/

#include <strstream.h>
#include <cstring>
#include <stdio.h>

#include <Dal.h>
#include <Region.h>
#include <dss.h>
#include <errstr.h>


template <class T>
DSrangeFilter<T>::DSrangeFilter(DScomp *parent, const string &name, 
				const list<Range<T> > ranges,
				const string &units) :
  DSfilter(parent, RangeT, name, units), _ranges(ranges) {}


template <class T>
DSrangeFilter<T>::DSrangeFilter(DScomp *parent,
				const Column *inheritedColumn,
				const list<Range<T> > ranges) :
  DSfilter(parent, RangeT, inheritedColumn),  _ranges(ranges) {}


template <class T>
DSintRange<T>::DSintRange(DScomp *parent,
			  const string &name, const list<Range<T> > ranges,
			  const string &units) :
  DSrangeFilter<T>(parent, name, ranges, units) {}


template <class T>
DSrealRange<T>::DSrealRange(DScomp *parent,
			    const string &name, const list<Range<T> > ranges,
			    const string &units) :
  DSrangeFilter<T>(parent, name, ranges, units) {}


template <class T>
DSintRange<T>::DSintRange(DScomp *parent,
			  const Column *inheritedColumn,
			  const list<Range<T> > ranges) :
  DSrangeFilter<T>(parent, inheritedColumn, ranges) {

    // Make sure the Column type is compatible

    Dal::DataType colType = inheritedColumn->dataType();

    if( colType != Dal::Int8 && 
	colType != Dal::Int16 && 
	colType != Dal::Int32 ){
      errstr << "Incompatible data types for DSS and column specifications"
	     << Sas::error("DSStype");
    }
}


template <class T>
DSrealRange<T>::DSrealRange(DScomp *parent,
			    const Column *inheritedColumn,
			    const list<Range<T> > ranges) :
  DSrangeFilter<T>(parent, inheritedColumn, ranges) {

    // Make sure the Column type is compatible

    Dal::DataType colType = inheritedColumn->dataType();

    if( colType != Dal::Real32 && 
	colType != Dal::Real64 ){
      errstr << "Incompatible data types for DSS and column specifications"
	     << Sas::error("DSStype");
    }
}


template <class T>
list<Range<double> > DSrangeFilter<T>::ranges() const{
  list<Range<double> > r;
  list<Range<T> >::const_iterator irange;

  for(irange=_ranges.begin(); irange!=_ranges.end(); irange++){
    Range<double> trange(*irange);
    r.push_back(trange);
  }

  return r;
}


template <class T> 
Range<double> DSrangeFilter<T>::extent(const string &axis) const{
  if( axis!=_name )
    errstr << "Attempt to extract extent information from the wrong filter"
	   << Sas::error("DSextentRange");

  Range<T> r = maxRange(_ranges);
  return r;
}


template <class T>
string DSrangeFilter<T>::selectExpression() {
  list<Range<T> >::const_iterator irange;
  strstream expression;
  T upper, lower;
  RangeBound lowerB, upperB;

  expression << name() << " in ";

  for(irange=_ranges.begin(); irange!=_ranges.end(); irange++){
    (*irange).values(lower, upper, lowerB, upperB);

    if( lowerB==Closed ) expression << "[";
    else expression << "(";

    if( lowerB!=Unbounded ){
      expression.setf(ios::showpoint);
      expression.precision(15);
      expression << lower;
    }

    if( lowerB==Unbounded || upperB==Unbounded || lower!=upper ) expression << ":";

    if( upperB!=Unbounded && (lowerB==Unbounded||lower!=upper) ){
      expression.setf(ios::showpoint);
      expression.precision(15);
      expression << upper;
    }

    if( upperB==Closed ) expression << "]";
    else expression << ")";

    if( irange != --_ranges.end() ) expression << ",";
  }

  expression << ends;
  return expression.str();
}


template <class T>
void DSrangeFilter<T>::write(unsigned compNumber,  unsigned filterNumber,
			     Block *toBlock, bool dummy) const{

  // Begin by making the range specification string

  strstream expression;

  if( dummy )
    expression << ":";
  else{
    list<Range<T> >::const_iterator irange;
    T upper, lower;
    RangeBound lowerB, upperB;

    for(irange=_ranges.begin(); irange!=_ranges.end(); irange++){
      (*irange).values(lower, upper, lowerB, upperB);

      if( lowerB==Open ) expression << "(";

      if( lowerB!=Unbounded ){
	expression.setf(ios::showpoint);
	expression.precision(15);
	expression << lower;
      }

      if( lowerB==Unbounded || upperB==Unbounded || lower!=upper ) expression << ":";

      if( upperB!=Unbounded && (lowerB==Unbounded||lower!=upper) ){
	expression.setf(ios::showpoint);
	expression.precision(15);
	expression << upper;
      }

      if( upperB==Open ) expression << ")";

      if( irange != --_ranges.end() ) expression << ",";
    }
  }

  expression << ends;

  // Now fill in the attributes

  strstream valueName;
  
  if( compNumber==1 )
    valueName << "DSVAL" << filterNumber << ends;
  else
    valueName << compNumber << "DSVAL" << filterNumber << ends;

  *toBlock->addAttribute(valueName.str(), comment+"value") = 
    expression.str();
}


template<class T>
double DSrealRange<T>::volume() const{
  list<Range<T> >::const_iterator irange;
  double vol=0;
  T upper, lower;
  RangeBound lowerB, upperB;

  for(irange=_ranges.begin(); irange!=_ranges.end(); irange++){
    (*irange).values(lower, upper, lowerB, upperB);
    
    if( lowerB==Unbounded || upperB==Unbounded ){
      errstr << "Cannot calculate volume of a range specification "
	     << "with only one limit.  Zero volumed returned."
             << Sas::warning("RangeVol");
      return 0;
    }

    vol += upper-lower;
  }

  return vol;
}


template<class T>
double DSintRange<T>::volume() const{
  list<Range<T> >::const_iterator irange;
  double vol=0;
  T upper, lower;
  RangeBound lowerB, upperB;

  for(irange=_ranges.begin(); irange!=_ranges.end(); irange++){
    irange->values(lower, upper, lowerB, upperB);
    
    if( lowerB==Unbounded || upperB==Unbounded ){
      errstr << "Cannot calculate volume of a range specification "
	     << "with only one limit.  Zero volumed returned."
             << Sas::warning("RangeVol");
      return 0;
    }

    vol += upper-lower+1;
    if( lowerB==Open ) vol -= 1;
    if( upperB==Open ) vol -= 1;
  }

  return vol;
}


template<class T>
bool DSrangeFilter<T>::operator==(DSfilter & oFilter) const{

  if( name() != oFilter.name()   ||
      type() != oFilter.type()   ||
      units() != oFilter.units() ||
      comps() != oFilter.comps() )
    return false;
      
  DSrangeFilter<T> * oDSrange = dynamic_cast<DSrangeFilter<T> *>(&oFilter);
  if( oDSrange == 0 ) return false;

  if( ranges() != oDSrange->ranges() ) return false;

  return true;
}

template class DSrangeFilter<long>;
template class DSrangeFilter<double>;

template class DSintRange<long>;
template class DSrealRange<double>;
