//
//  Implementation of the DSrangeBLock class
//
//	AUTHOR:
//	  D. Hinshaw (DH)
//
//	  REFERENCES:
//	    [1] FITS Keyword Conventions in ASC Data Model files, SDS-7.2
// 		J. McDowell
//		http://hea-www.harvard.edu/~jcm/asc/docs/
//
//          [2] ASC Data Model Abstact Design, SDS-3.0
//              J. McDowell
//		http://hea-www.harvard.edu/~jcm/asc/docs/
//	    
//	    [3] Filtering Syntax in the ASC Data Model, SDS-6.0
//	        J. McDowell, A. Fruscione, A. Siemiginowska, B. Joye
//	        http://hea-www.harvard.edu/~jcm/asc/docs/

#include <strstream.h>
#include <cstring>

#include <Dal.h>
#include <Region.h>
#include <dss.h>
#include <errstr.h>


DSrangeBlock::DSrangeBlock(DSrangeBlock &dsblock) : DSblockFilter(dsblock) {}


DSrangeBlock::DSrangeBlock(DScomp *parent,
			   const string &name, const string &blockSpec,
			   const string &units = "") :
  DSblockFilter(parent, RangeBlockT, name, blockSpec, units){

    // Check that this block contains a compatible range table

    if( _block->type() == Block::TableT ){
      for(unsigned i=1;;i++){
	strstream mtype;
	mtype << "MTYPE" << i << ends;

	if( _block->hasAttribute(mtype.str()) ){
	  if( _block->attribute(mtype.str())->asString() == _name ) return;
	}else{
	  break;
	}
      }
    }

    errstr << "Block type is not compatible with the range filter type"
	   << Sas::error("BlockTypeRange");
    return;
}


DSrangeBlock::DSrangeBlock(DScomp *parent,
			   const Column *inheritedColumn, const string &blockSpec) :
  DSblockFilter(parent, RangeBlockT, inheritedColumn, blockSpec){

    // Check that this block contains a compatible range table

    if( _block->type() == Block::TableT ){
      for(unsigned i=1;;i++){
	strstream mtype;
	mtype << "MTYPE" << i << ends;

	if( _block->hasAttribute(mtype.str()) ){
	  if( _block->attribute(mtype.str())->asString() == name() ) return;
	}else{
	  break;
	}
      }
    }

    errstr << "Block type is not compatible with the range filter type"
	   << Sas::error("BlockTypeRange");
    return;
}



string DSrangeBlock::selectExpression() {
  strstream expression;
  unsigned i;

  expression << name() << " in ";

  for(i=1;;i++){
    strstream mtype;
    mtype << "MTYPE" << i << ends;
    if( _block->hasAttribute(mtype.str()) ){
      if( _block->attribute(mtype.str())->asString() == name() ) break;
    }
  }

  strstream mform;
  mform << "MFORM" << i << ends;
  string forms(_block->attribute(mform.str())->asString());
  int icomma = forms.find(",");

  string lowerName = forms.substr(0,icomma);
  string upperName = forms.substr(icomma+1);

  Table * table = dynamic_cast<Table *>(_block);

  ColumnData *lowerData = table->column(lowerName)->data();
  ColumnData *upperData = table->column(upperName)->data();

  long unsigned nrows = table->rows();


  #define CODE(type)                                          \
  {                                                           \
    type *lowerVal = lowerData->type##Data();                 \
    type *upperVal = upperData->type##Data();                 \
                                                              \
    for(long unsigned irow=0; irow<nrows; irow++){            \
      expression.setf(ios::showpoint);                        \
      expression << lowerVal[irow];                           \
      expression << ":";                                      \
      expression.setf(ios::showpoint);                        \
      expression << upperVal[irow];                           \
      if(irow!=nrows-1) expression << ",";                    \
    }                                                         \
    break;                                                    \
  }                                                           \

  switch(table->column(lowerName)->dataType()){

  case Dal::Int32  : CODE(int32)
		      
  case Dal::Int16  : CODE(int16)

  case Dal::Int8   : CODE(int8)

  case Dal::Real32 : CODE(real32)

  case Dal::Real64 : CODE(real64)

  default :
    errstr << "Columns are not of a supported data type." << Sas::error("RangeType");
  }

  #undef CODE


  expression <<ends;

  delete lowerData;
  delete upperData;

  return expression.str();
}


Range<double> DSrangeBlock::extent(const string &axis) const{
  if( axis!=_name )
    errstr << "Attempt to extract extent information from the wrong filter"
	   << Sas::error("DSextentRange");

  unsigned i;
  for(i=1;;i++){
    strstream mtype;
    mtype << "MTYPE" << i << ends;
    if( _block->hasAttribute(mtype.str()) ){
      if( _block->attribute(mtype.str())->asString() == name() ) break;
    }
  }

  strstream mform;
  mform << "MFORM" << i << ends;
  string forms(_block->attribute(mform.str())->asString());
  int icomma = forms.find(",");

  string lowerName = forms.substr(0,icomma);
  string upperName = forms.substr(icomma+1);

  Table * table = dynamic_cast<Table *>(_block);

  ColumnData *lowerData = table->column(lowerName)->data();
  ColumnData *upperData = table->column(upperName)->data();

  long unsigned nrows = table->rows();

  list<Range<double> > ranges;

  #define CODE(type)                                          \
  {                                                           \
    type *lowerVal = lowerData->type##Data();                 \
    type *upperVal = upperData->type##Data();                 \
                                                              \
    for(long unsigned irow=0; irow<nrows; irow++){            \
      Range<double> r(lowerVal[irow], upperVal[irow]);        \
      ranges.push_back(r);                                    \
    }                                                         \
    break;                                                    \
  }                                                           \

  switch(table->column(lowerName)->dataType()){

  case Dal::Int32  : CODE(int32)
		      
  case Dal::Int16  : CODE(int16)

  case Dal::Int8   : CODE(int8)

  case Dal::Real32 : CODE(real32)

  case Dal::Real64 : CODE(real64)

  default :
    errstr << "Columns are not of a supported data type." << Sas::error("RangeType");
  }

  #undef CODE

  delete lowerData;
  delete upperData;

  Range<double> result = maxRange(ranges);
  return result;
}


double DSrangeBlock::volume() const{
  double vol=0;

  unsigned i;

  for(i=1;;i++){
    strstream mtype;
    mtype << "MTYPE" << i << ends;
    if( _block->hasAttribute(mtype.str()) ){
      if( _block->attribute(mtype.str())->asString() == name() ) break;
    }
  }

  strstream mform;
  mform << "MFORM" << i << ends;
  string forms(_block->attribute(mform.str())->asString());
  int icomma = forms.find(",");

  string lowerName = forms.substr(0,icomma);
  string upperName = forms.substr(icomma+1);

  Table * table = dynamic_cast<Table *>(_block);

  ColumnData *lowerData = table->column(lowerName)->data();
  ColumnData *upperData = table->column(upperName)->data();

  long unsigned nrows = table->rows();


  #define CODEREAL(type)                                      \
  {                                                           \
    type *lowerVal = lowerData->type##Data();                 \
    type *upperVal = upperData->type##Data();                 \
                                                              \
    for(long unsigned irow=0; irow<nrows; irow++){            \
      vol += upperVal[irow]-lowerVal[irow];                   \
    }                                                         \
    break;                                                    \
  }                                                           \


  #define CODEINT(type)                                       \
  {                                                           \
    type *lowerVal = lowerData->type##Data();                 \
    type *upperVal = upperData->type##Data();                 \
                                                              \
    for(long unsigned irow=0; irow<nrows; irow++){            \
      vol += upperVal[irow]-lowerVal[irow]+1;                 \
    }                                                         \
    break;                                                    \
  }                                                           \

  switch(table->column(lowerName)->dataType()){

  case Dal::Int32  : CODEINT(int32)
		      
  case Dal::Int16  : CODEINT(int16)

  case Dal::Int8   : CODEINT(int8)

  case Dal::Real32 : CODEREAL(real32)

  case Dal::Real64 : CODEREAL(real64)

  default :
    errstr << "Columns are not of a supported data type." << Sas::error("RangeType");
  }

  #undef CODEREAL
  #undef CODEINT

  delete lowerData;
  delete upperData;

  return vol;  
}
