//
//  Implementation of the DSregionFilter class
//
//	AUTHOR:
//	  D. Hinshaw (DH)
//
//	  REFERENCES:
//	    [1] FITS Keyword Conventions in ASC Data Model files, SDS-7.2
// 		J. McDowell
//		http://hea-www.harvard.edu/~jcm/asc/docs/
//
//          [2] ASC Data Model Abstact Design, SDS-3.0
//              J. McDowell
//		http://hea-www.harvard.edu/~jcm/asc/docs/
//	    
//	    [3] Filtering Syntax in the ASC Data Model, SDS-6.0
//	        J. McDowell, A. Fruscione, A. Siemiginowska, B. Joye
//	        http://hea-www.harvard.edu/~jcm/asc/docs/

#include <strstream.h>
#include <cstring>
#include <math.h>
#include <cstdio>

#include <Dal.h>
#include <Region.h>
#include <ASCregion.h>
#include "DataFilter.h"
#include <dss.h>
#include <errstr.h>


DSregionFilter::DSregionFilter(DScomp *parent,
			       const string &name, GeoRegion region,
			       DSfilterComp &filtercomp) :
  DSfilter(parent, RegionT, name, "", filtercomp), _region(region) {}


DSregionFilter::DSregionFilter(DScomp *parent,
			       const string &name, GeoRegion region) :
  DSfilter(parent, RegionT, name), _region(region) {}


GeoRegion DSregionFilter::region() const{
  return _region;
}


void DSregionFilter::write(unsigned compNumber, unsigned filterNumber, 
			   Block *toBlock, bool dummy) const{

  // A keyword region specification is pretty limited, so instead
  // write out region filters as references to FITS region blocks.

  strstream valueName0, refName, ref;
  
  ref << "REG" << compNumber*100 + filterNumber << ends;

  valueName0 << "DSVAL" << filterNumber << ends;

  if( !toBlock->hasAttribute(valueName0.str()) ||
      toBlock->attribute(valueName0.str())->asString() != "TABLE" )
    *toBlock->addAttribute(valueName0.str(), comment+"value") = "TABLE";

  DSfilterComp filterComps = comps();
  strstream typeName;
  typeName << "DSTYP" << filterNumber << ends;
    
  *toBlock->attribute(typeName.str()) = toBlock->attribute(typeName.str())->asString() + 
    "(" + filterComps[0] + "," + filterComps[1] + ")";

  strstream refValue;
  if( dummy )
    refValue << ":" << ends;
  else
    refValue << ":" << ref.str() << ends;

  strstream refName0;
  refName0 << "DSREF" << filterNumber << ends;

  if( compNumber==1 )
    *toBlock->addAttribute(refName0.str(), comment+"reference")
      = refValue.str();
  else if( ! toBlock->hasAttribute(refName0.str()) ||
	   toBlock->attribute(refName0.str())->asString() != refValue.str() ){
    strstream refName;
    refName << compNumber << "DSREF" << filterNumber << ends;
    *toBlock->addAttribute(refName.str(), comment+"reference") 
      = refValue.str();
  }

  // Need to copy the region to the output data set.
  // First check for a name clash.

  if( !dummy ){
  
    if( toBlock->parent()->hasBlock(ref.str()) )
      errstr << "Name clash while writing Data Subspace: Block "
	     << ref.str() << " already exists in output data set."
	     << Sas::error("DssBlockClash");

    // Now write the region table

    Table *ascTable = toBlock->parent()
                        ->addTable(ref.str(), _region.elements());
    DSfilterComp regionComps = comps();

    OutASCregionTable ascRegion(ascTable, regionComps[0], regionComps[1]);
    ascRegion << _region;
  }
}



string DSregionFilter::selectExpression() {
  strstream expression;

  DSfilterComp filtComps = comps();

  for (unsigned comp=0; comp<_region.size(); ++comp){
    expression << "(";

    for (unsigned element=0; element<_region[comp].size(); ++element) {
      if(! _region[comp][element]->included() ) expression << "!";

      switch(_region[comp][element]->type()){
      case ShapeBasics::POINT         : {
	expression << "point(";
	break;
      }
      case ShapeBasics::CIRCLE        : {
	expression << "circle(";
	break;
      }
      case ShapeBasics::ANNULUS       : {
	expression << "ring(";
	break;
      }
      case ShapeBasics::ELLIPSE       : {
	expression << "ellipse(";
	break;
      }
      case ShapeBasics::ELLIPTANNULUS : {
	expression << "elliptring(";
	break;
      }
      case ShapeBasics::SECTOR        : {
	expression << "sector(";
	break;
      }
      case ShapeBasics::BOX           : {
	expression << "box(";
	break;
      }
      case ShapeBasics::ROTBOX        : {
	expression << "box(";
	break;
      }
      case ShapeBasics::RECTANGLE     : {
	expression << "rectangle(";
	break;
      }
      case ShapeBasics::ROTRECTANGLE  : {
	expression << "rectangle(";
	break;
      }
      case ShapeBasics::RHOMBUS       : {
	expression << "diamond(";
	break;
      }
      case ShapeBasics::ROTRHOMBUS    : {
	expression << "diamond(";
	break;
      }
      case ShapeBasics::POLYGON       : {
	expression << "polygon(";
	break;
      }
      case ShapeBasics::LINE          : {
	expression << "line(";
	break;
      }
      }

      vector<float> X = _region[comp][element]->pointsX();
      vector<float> Y = _region[comp][element]->pointsY();
      vector<float> R = _region[comp][element]->radii();
      vector<double> angles = _region[comp][element]->rotAngles();

      for(unsigned point=0; point<X.size(); point++){
	expression << X[point] << "," << Y[point];
	if( point != X.size()-1 ) expression << ",";
      }

      if( R.size() !=0 ) expression << ",";

      for(unsigned point=0; point<R.size(); point++){
	expression << R[point];
	if( point != R.size()-1 ) expression << ",";
      }

      if( angles.size() == 0 ){
	if( _region[comp][element]->type()==ShapeBasics::BOX ||
	    _region[comp][element]->type()==ShapeBasics::RECTANGLE ||
	    _region[comp][element]->type()==ShapeBasics::RHOMBUS )
	  expression << ",0";

      }else{
	expression << ",";
	
	for(unsigned point=0; point<angles.size(); point++){
	  expression << angles[point];
	  if( point != angles.size()-1 ) expression << ",";
	}
      }

      expression << "," << filtComps[0] << "," << filtComps[1] << ")";
      
      if( element != _region[comp].size()-1 ) expression << "&&";
    }

    expression << ")";
    if( comp != _region.size()-1 ) expression << "||";
  }

  expression << ends;
  return expression.str();
}


Range<double> DSregionFilter::extent(const string &axis) const{
  int Xmin, Xmax, Ymin, Ymax;
  geoExtent(_region, Xmin, Xmax, Ymin, Ymax);

  DSfilterComp lcomps = comps();
  if( axis==lcomps[0] ){
    Range<double> r(Xmin, Xmax);
    return r;
  }else if( axis==lcomps[1] ){
    Range<double> r(Ymin, Ymax);
    return r;
  }

  errstr << "Attempt to extract extent information from the wrong filter"
	 << Sas::error("DSextentRegion");

  Range<double> r(0, 0, Unbounded, Unbounded);
  return r;
}


double DSregionFilter::volume() const{
  return geoArea(_region);
}


bool DSregionFilter::operator==(DSfilter &oFilter) const{

  if( name() != oFilter.name()   ||
      type() != oFilter.type()   ||
      units() != oFilter.units() ||
      comps() != oFilter.comps() )
    return false;
      
  if( region() != oFilter.region() ) return false;

  return true;
}

void geoExtent(const GeoRegion &region, int &Xmin, int &Xmax, int &Ymin, int &Ymax){

  for (unsigned comp=0; comp<region.size(); ++comp){
    for (unsigned element=0; element<region[comp].size(); ++element) {
      vector<float> X = region[comp][element]->pointsX();
      vector<float> Y = region[comp][element]->pointsY();
      vector<float> R = region[comp][element]->radii();
      float Rmax = 0;
      for(unsigned ir=0; ir<R.size(); ir++) 
	if( R[ir]>Rmax ) Rmax = R[ir];
      if(Rmax==0)
	for(unsigned i=1; i<X.size(); i++){
	  float iR = sqrt((X[0]-X[i])*(X[0]-X[i])+(Y[0]-Y[i])*(Y[0]-Y[i]));
	  if( iR>Rmax ) Rmax = iR;
	}
      if(comp==0 && element==0){
	Xmin = (int)floor(X[0]-Rmax);
	Xmax = (int)ceil(X[0]+Rmax);
	Ymin = (int)floor(Y[0]-Rmax);
	Ymax = (int)ceil(Y[0]+Rmax);
      }else{
	if( floor(X[0]-Rmax) < Xmin ) Xmin = (int)floor(X[0]-Rmax);
	if( ceil(X[0]+Rmax) > Xmax )  Xmax = (int)ceil(X[0]+Rmax);
	if( floor(Y[0]-Rmax) < Ymin ) Ymin = (int)floor(Y[0]-Rmax);
	if( ceil(Y[0]+Rmax) > Ymax )  Ymax = (int)ceil(Y[0]+Rmax);
      }
    }
  }

}

unsigned long geoArea(const GeoRegion &region){
  int Xmin, Xmax, Ymin, Ymax;

  geoExtent(region, Xmin, Xmax, Ymin, Ymax);

  DataSet *tempSet = dataSetServer->open(tempnam(0,0), Dal::Temp);
  Table *tempTab = tempSet->addTable("GRID", (Xmax-Xmin+1)*(Ymax-Ymin+1));
  ColumnData *Xdata = tempTab->addColumn("X", Dal::Int16)->data();
  ColumnData *Ydata = tempTab->addColumn("Y", Dal::Int16)->data();
  
  int16 *X = Xdata->int16Data();
  int16 *Y = Ydata->int16Data();

  unsigned long count = 0;
  for(int ix=Xmin; ix<=Xmax; ix++){
    for(int iy=Ymin; iy<=Ymax; iy++){
      X[count] = ix;
      Y[count] = iy;
      count++;
    }
  }

  delete Xdata;
  delete Ydata;

  DSregionFilter rFilter(NULL, "POS", region);
  DataFilter filter(rFilter.selectExpression());
  vector<DataFilter::RowRange> rowranges;

  unsigned long area = filter.doSelect(*tempTab, rowranges);
  dataSetServer->close(tempSet);

  return area;
}
