//
//  IMplementation of the DataSubspace class
//
//	AUTHOR:
//	  D. Hinshaw (DH)
//
//	  REFERENCES:
//	    [1] FITS Keyword Conventions in ASC Data Model files, SDS-7.2
// 		J. McDowell
//		http://hea-www.harvard.edu/~jcm/asc/docs/
//
//          [2] ASC Data Model Abstact Design, SDS-3.0
//              J. McDowell
//		http://hea-www.harvard.edu/~jcm/asc/docs/
//	    
//	    [3] Filtering Syntax in the ASC Data Model, SDS-6.0
//	        J. McDowell, A. Fruscione, A. Siemiginowska, B. Joye
//	        http://hea-www.harvard.edu/~jcm/asc/docs/


#include <stdio.h>
#include <strstream.h>
#include <cstring>
#include <map>


#include <Dal.h>
#include <Region.h>
#include <DataFilter.h>
#include <dss.h>
#include <errstr.h>


DSS::DSaxes const DSS::noAxes(0);


DataSubspace::DataSubspace() : _compvec(0) {}


DataSubspace::DataSubspace(const DataSubspace &dss){

  for(unsigned icomp=0; icomp<dss.components(); icomp++){
    DScomp * comp = new DScomp( *dss.component(icomp) );
    comp->parent(this);
    _compvec.push_back(comp);
  }

  return;
}


DataSubspace &DataSubspace::operator=(const DataSubspace &dss){
  vector<DScomp *>::iterator icomp;
  for(icomp=_compvec.begin(); icomp!=_compvec.end(); icomp++)
    delete *icomp;
  _compvec.clear();

  for(unsigned icomp=0; icomp<dss.components(); icomp++){
    DScomp * comp = new DScomp( *dss.component(icomp) );
    comp->parent(this);
    _compvec.push_back(comp);
  }

  return *this;
}


DataSubspace::DataSubspace(Block *inBlock) : _compvec(0){
  inBlock >> *this;
}


DScomp *DataSubspace::addComponent(){
  DScomp *comp = new DScomp(this);
  _compvec.push_back(comp);

  return comp;
}


DScomp *DataSubspace::addComponent(const DScomp &dscomp){
  DScomp *comp = new DScomp(dscomp);
  comp->parent(this);

  _compvec.push_back(comp);
  return comp;
}


void DataSubspace::deleteComponent(unsigned compNumber){
  delete *( _compvec.begin()+compNumber );
  _compvec.erase( _compvec.begin()+compNumber );
}


DScomp *DataSubspace::component(unsigned compNumber) const{
  return _compvec[compNumber];
}


unsigned DataSubspace::components() const{
  return _compvec.size();
}


string DataSubspace::selectExpression(const DSaxes &axes) const{
  vector<DScomp * const>::iterator icomp;
  strstream expression;

  bool firstcomp = true;
  for(icomp=_compvec.begin(); icomp!=_compvec.end(); icomp++){
    string compstr = (*icomp)->selectExpression(axes);
    if( compstr.size() != 0 ){
      if(firstcomp) firstcomp=false;
      else expression << " || ";
      expression << "( " << compstr << " )";
    }
  }

  expression << ends;

  return expression.str();
}


Range<double> DataSubspace::extent(const string &axis) const{
  vector<DScomp * const>::iterator icomp;
  list<Range<double> > rlist;

  for(icomp=_compvec.begin(); icomp!=_compvec.end(); icomp++)
    rlist.push_back( (*icomp)->extent(axis) );

  return maxRange(rlist);
}


double DataSubspace::volume(const DSaxes &axes) const{
  vector<DScomp * const>::iterator icomp;
  double vol=0;

  for(icomp=_compvec.begin(); icomp!=_compvec.end(); icomp++)
    vol += (*icomp)->volume(axes);

  return vol;
}


list<Range<double> > DataSubspace::ranges(const string &axis) const{
  vector<DScomp *>::const_iterator icomp;
  list<Range<double> > r;

  for(icomp=_compvec.begin(); icomp!=_compvec.end(); icomp++){
    list<Range<double> > rtemp = (*icomp)->ranges(axis);
    r.merge(rtemp);
  }

  return r;
}


bool DataSubspace::hasFilter(const DSfilter &filter) const{
  vector<DScomp *>::const_iterator icomp;

  for(icomp=_compvec.begin(); icomp!=_compvec.end(); icomp++){
    if( (*icomp)->hasFilter(filter) ) return true;
  }  

  return false;
}


bool DataSubspace::isGlobal(const DSfilter &filter) const{

  for(unsigned icomp=0; icomp<components(); icomp++){
    DScomp *comp = component(icomp);
    bool global = false;
    for(unsigned ifilter=0; ifilter<comp->filters(); ifilter++)
      if( filter == *comp->filter(ifilter) ){
	global = true;
	break;
      }
	
    if(!global) return false;
  }

  return true;
}


void DataSubspace::compressRegions() {
  vector<DScomp *>::const_iterator icomp;

  for(icomp=_compvec.begin(); icomp!=_compvec.end(); icomp++)
    (*icomp)->compressRegions();

  return;
}


DataSubspace::~DataSubspace(){
  vector<DScomp *>::iterator icomp;

  for(icomp=_compvec.begin(); icomp!=_compvec.end(); icomp++)
    delete *icomp;

  return;
}


Block *operator<<(Block *toBlock, const DataSubspace &fromDss){

  // Check to see if the block already contains DSS info,
  // if it does, read it in and merge it with the current DSS,
  // before writing it back out.

  DataSubspace oldDss;
  DataSubspace outDss(fromDss);

  if( fromDss.components() == 0 ) return toBlock;

  string checkType = "DSTYP1";
  if( toBlock->hasAttribute(checkType) ){
    toBlock >> oldDss;
    deleteDSS(toBlock); 
    outDss *= oldDss;
  }

  DScomp * comp;
  DSfilter * filter;
  unsigned icomp, ifilt;
  string comment = "data subspace descriptor: ";

  vector< DSfilter *> attrFilters;

  // Make a first pass thru the DSS, writing type and unit 
  // keywords and getting a list of the filters

  for(icomp=0; icomp < outDss.components(); icomp++){
    comp = outDss.component(icomp);
    for(ifilt=0; ifilt < comp->filters(); ifilt++){
      filter = comp->filter(ifilt);

      // See if this filter has already been defined,
      // if not, add it to the list.

      bool inList = false;
      for( unsigned i=0; i<attrFilters.size(); i++)
	if( filter->like(*attrFilters[i]) ){
	  inList = true;
	  break;
	}

      if( !inList ){
	attrFilters.push_back(filter);

	// Check for a name clash between this filter and a column

	if( toBlock->type() == Block::TableT ){
	  Table * toTable = dynamic_cast<Table *>(toBlock);
	  if( toTable->hasColumn(filter->name()) &&
	      ( filter->type() == DSS::RegionT ||
		filter->type() == DSS::MaskT ||
		filter->type() == DSS::RegionBlockT ||
		filter->units() != toTable->column(filter->name())->units() )
	      )
	    errstr << "Data subspace filter name and column name "
		   << filter->name()
		   << " are not consistent.  Cannot write data subspace "
		   << "to table " << toTable->name() << Sas::error("ColumnClash");
	}

	strstream name0;
	name0 << "DSTYP" << attrFilters.size()  << ends;
	*toBlock->addAttribute(name0.str(), comment+"name") = filter->name();

	if( filter->units() != "" ){
	  strstream unit0;
	  unit0 << "DSUNI" << attrFilters.size() << ends;
	  *toBlock->addAttribute(unit0.str(), comment+"units") = 
	    filter->units();
	}


      }	
    }
  }

  // Now make a second pass thru the DSS, writing value and
  // reference keywords, as appopriate for the type of filter.  

  DScomp * comp0 = outDss.component(0);
  for(icomp=0; icomp < outDss.components(); icomp++){
    comp = outDss.component(icomp);
    for( unsigned ifilt=0; ifilt<attrFilters.size(); ifilt++){

      if( comp->hasFilter(*attrFilters[ifilt]) ){
	if( icomp==0 ||
	    !comp0->hasFilter(*attrFilters[ifilt]) ||
	    *comp0->filter(*attrFilters[ifilt]) != 
	    *comp->filter(*attrFilters[ifilt]) )

	  comp->filter(*attrFilters[ifilt])
	    ->write(icomp+1, ifilt+1, toBlock);

      }else if( icomp==0 || comp0->hasFilter(*attrFilters[ifilt]) ){

	// No filter for this component, so write a dummy
	attrFilters[ifilt]
	  ->write(icomp+1, ifilt+1, toBlock, true);

      }
    }
  }

  return toBlock;
}


Block *operator>>(Block *fromBlock, DataSubspace &toDss){
  
  DataSubspace inDss;

  // Count the number of DSS filters and components in the block
  
  unsigned nFilters;
  
  for(nFilters=0; ; nFilters++){
    strstream typeName;
    typeName << "DSTYP" << nFilters+1 << ends;
    if( ! fromBlock->hasAttribute(typeName.str()) ) break;
  }
  
  if( nFilters==0 ) return fromBlock;
  
  unsigned nComps;
  
  for(nComps=1; ; nComps++){
    inDss.addComponent();

    bool values = false;
    for(unsigned ifilt=0; ifilt<nFilters; ifilt++){
      strstream valName, refName;
      valName << nComps+1 << "DSVAL" << ifilt+1 << ends;
      refName << nComps+1 << "DSREF" << ifilt+1 << ends;
      if( fromBlock->hasAttribute(valName.str()) ||
 	  fromBlock->hasAttribute(refName.str()) )
 	values = true;
    }
    
    if( !values )break;
  }
  
  // Now loop thru again, adding filters as needed
  
  map<string, string> filterBlocks;

  for(unsigned ifilt=1; ifilt<=nFilters; ifilt++){
    strstream typeName, unitName;
    typeName << "DSTYP" << ifilt << ends;

    bool hasFiltComp = false;
    string compVal;
    string typeVal = fromBlock->attribute(typeName.str())->asString();
    string::size_type iparse = typeVal.find("(");
    if( iparse != string::npos ){
      hasFiltComp = true;
      compVal = typeVal.substr(iparse);
      typeVal = typeVal.substr(0,iparse);
    }

    unitName << "DSUNI" << ifilt << ends;
    string unitVal;
    if( fromBlock->hasAttribute(unitName.str()) )
      unitVal = fromBlock->attribute(unitName.str())->asString();
    else
      unitVal = "";

    // Is there a column associated with this filter?

    Column * inheritCol = 0;
    bool inherit = false;
    if( fromBlock->type() == Block::TableT ){
      Table * fromTable = dynamic_cast<Table *>(fromBlock);
      if( fromTable->hasColumn(typeVal) ){
	inherit = true;
	inheritCol = fromTable->column(typeVal);
      }
    }

    for(unsigned icomp=0; icomp<nComps; icomp++){
      strstream valName0;
      valName0 << "DSVAL" << ifilt << ends;
      string value;
      
      // Get the filter value

      if( icomp==0 ){
	value = fromBlock->attribute(valName0.str())->asString();
      }else{
	strstream valName;
	valName << icomp+1 << "DSVAL" << ifilt << ends;
	if( fromBlock->hasAttribute(valName.str()) )
	  value = fromBlock->attribute(valName.str())->asString();
	else
	  value = fromBlock->attribute(valName0.str())->asString();
      }

      // Determine filter type.  Do nothing if empty filter

      if( value==":" ) continue;

      // Block filter

      if( value=="IMAGE" || value=="TABLE" ){
	strstream refName0;
	refName0 << "DSREF" << ifilt << ends;
	string ref;
	
	if( icomp==0 ){
	  ref = fromBlock->attribute(refName0.str())->asString();
	}else{
	  strstream refName;
	  refName << icomp+1 << "DSREF" << ifilt << ends;
	  if( fromBlock->hasAttribute(refName.str()) )
	    ref = fromBlock->attribute(refName.str())->asString();
	  else
	    ref = fromBlock->attribute(refName0.str())->asString();
	}

	if( ref==":" ) continue;
	ref.erase(0,1);
	
	if( filterBlocks[ref]=="" )
	  filterBlocks[ref] = fromBlock->parent()->name() + ":" + ref;

	if(inherit){
	  inDss.component(icomp)->addFilter(inheritCol, filterBlocks[ref]);
	  continue;
	}

	if( hasFiltComp ){
	  DSS::DSfilterComp dscomps;
	  string::size_type icomma = compVal.find(",");
	  string::size_type oparen = compVal.find("(");
	  string::size_type cparen = compVal.find(")");

	  dscomps.push_back(compVal.substr(oparen+1,icomma-oparen-1));
	  dscomps.push_back(compVal.substr(icomma+1,cparen-icomma-1));

	  inDss.component(icomp)->addFilter(typeVal, filterBlocks[ref], dscomps);
	  continue;
	}

	inDss.component(icomp)->addFilter(typeVal, filterBlocks[ref], unitVal);
	continue;
      }

      // Bit filter
      
      strstream formName;
      formName << "DSFORM" << ifilt << ends;

      if( fromBlock->hasAttribute(formName.str()) &&
	  fromBlock->attribute(formName.str())->asString()=="X" ){

	unsigned long imask;
	sscanf(value.c_str(), "%li", &imask);

	if(inherit)
	  inDss.component(icomp)->addFilter(inheritCol, imask);
	else
	  inDss.component(icomp)->addFilter(typeVal, imask);

	continue;
      }

      // Range filter

      char firstVal = *value.substr(0,1).data();
      if( isdigit(firstVal) || firstVal=='[' ||
	  firstVal=='(' || firstVal=='.' || firstVal==':'){

	bool isInt = false;
	if( inherit )
	  isInt = inheritCol->dataType() == Dal::Int8 ||
	    inheritCol->dataType() == Dal::Int16 ||
	    inheritCol->dataType() == Dal::Int32;
	else if( fromBlock->hasAttribute(formName.str()) )
	  isInt = fromBlock->attribute(formName.str())->asString()=="B" ||
	    fromBlock->attribute(formName.str())->asString()=="I" ||
	    fromBlock->attribute(formName.str())->asString()=="J";
	else
	  isInt = value.find(".") == string::npos;
	
	if( isInt ){
	  list<Range<long> > inRanges = parseRanges<long>(value);
	  if(inherit) inDss.component(icomp)->addFilter(inheritCol, inRanges);
	  else inDss.component(icomp)->addFilter(typeVal, inRanges, unitVal);
	}else{
	  list<Range<double> > inRanges = parseRanges<double>(value);
	  if(inherit) inDss.component(icomp)->addFilter(inheritCol, inRanges);
	  else inDss.component(icomp)->addFilter(typeVal, inRanges, unitVal);
	}

	continue;
      }

      // Region filter - not supported

      errstr << "A data subspace filter with a region specification "
	     << "has been found.  This type of filter is not supported."
	     << Sas::error("BadRegionFilter");
    }

  }

  // If the DSS object is not empty, need to merge new DSS info with old.

  if( toDss.components() != 0 )
    toDss *= inDss;
  else
    toDss = inDss;

  return fromBlock;
}

// Perform union of two DSS's.  This simply involves ANDing together each of
// the components of one subspace with each of the components of the other.

DataSubspace &operator*=(DataSubspace &dss1, const DataSubspace &dss2){
  dss1 = dss1 * dss2;
  return dss1;
}


DataSubspace operator*(const DataSubspace &dss1, const DataSubspace &dss2){
  DataSubspace dssResult;
  
  for(unsigned icomp1=0; icomp1<dss1.components(); icomp1++)
    for(unsigned icomp2=0; icomp2<dss2.components(); icomp2++){
      dssResult.addComponent(*dss1.component(icomp1) * *dss2.component(icomp2));
    }

  return dssResult;
}
  

// Delete the DSS specification from a block, not including referenced extensions.
void deleteDSS(Block *block){

  // Count the number of DSS filters and components in the block
  
  unsigned nFilters;
  
  for(nFilters=0; ; nFilters++){
    strstream typeName;
    typeName << "DSTYP" << nFilters+1 << ends;
    if( ! block->hasAttribute(typeName.str()) ) break;

    block->deleteAttribute(typeName.str());
    
    strstream unitName;
    unitName << "DSUNI" << nFilters+1 << ends;
    if( block->hasAttribute(unitName.str()) ) block->deleteAttribute(unitName.str());
    
    strstream cmpName;
    cmpName << "DSCPT" << nFilters+1 << ends;
    if( block->hasAttribute(cmpName.str()) ) block->deleteAttribute(cmpName.str());
  }

  if( nFilters==0 ) return;
  
  for(unsigned nComps=1; ; nComps++){
    bool values = false;
    for(unsigned ifilt=0; ifilt<nFilters; ifilt++){
      strstream valName, refName;
      if( nComps==1 ){
	valName << "DSVAL" << ifilt+1 << ends;
	refName << "DSREF" << ifilt+1 << ends;
      }else{
	valName << nComps << "DSVAL" << ifilt+1 << ends;
	refName << nComps << "DSREF" << ifilt+1 << ends;
      }
      if( block->hasAttribute(valName.str()) ){
	block->deleteAttribute(valName.str());
	values = true;
      }
      if( block->hasAttribute(refName.str()) ){
	block->deleteAttribute(refName.str());
	values = true;
      }
    }
    if(!values) break; 
  }

}


void cleanDSS(const Table *tab, DataSubspace &dss){
  DSS::DSaxes cols;

  for(unsigned icol=0; icol<tab->columns(); icol++)
    cols.push_back(tab->column(icol)->name());

  for(unsigned icomp=dss.components(); icomp!=0; icomp--){
    DScomp * comp = dss.component(icomp-1);
    DSS::DSaxes axes = cols;

    for(unsigned ifilter=0; ifilter<comp->filters(); ifilter++){
      DSfilter * filter = comp->filter(ifilter);
      if( filter->type() != DSS::RegionT      &&
	  filter->type() != DSS::MaskT        &&
	  filter->type() != DSS::RegionBlockT ) continue;
      DSS::DSfilterComp filtComp = filter->comps();
      if( tab->hasColumn(filtComp[0]) && tab->hasColumn(filtComp[1]) )
	axes.push_back(filter->name());
    }

    DataFilter filter(comp->selectExpression(axes));
    vector<DataFilter::RowRange> rowranges;
    if( filter.doSelect(*tab, rowranges)==0 )
      dss.deleteComponent(icomp-1);
  }
}
