//
//  IMplementation of the Range class and associated utilities
//
//	AUTHOR:
//	  D. Hinshaw (DH)
//
//	  REFERENCES:
//	    [1] FITS Keyword Conventions in ASC Data Model files, SDS-7.2
// 		J. McDowell
//		http://hea-www.harvard.edu/~jcm/asc/docs/
//
//          [2] ASC Data Model Abstact Design, SDS-3.0
//              J. McDowell
//		http://hea-www.harvard.edu/~jcm/asc/docs/
//	    
//	    [3] Filtering Syntax in the ASC Data Model, SDS-6.0
//	        J. McDowell, A. Fruscione, A. Siemiginowska, B. Joye
//	        http://hea-www.harvard.edu/~jcm/asc/docs/

#include <math.h>
#include <strstream.h>
#include <cstring>

#include <dss.h>

template<class T> template<class U> Range<T>::Range(const Range<U> &inRange){
  U lower, upper;
  RangeBound lowerB, upperB;
  inRange.values(lower, upper, lowerB, upperB);

  _lower = static_cast<T>(lower);
  _upper = static_cast<T>(upper);
  _lowerB = lowerB;
  _upperB = upperB;  
}


template<class T> Range<T>::Range(T lower, T upper, 
				  RangeBound lowerB, RangeBound upperB) :
  _lower(lower), _upper(upper), _lowerB(lowerB), _upperB(upperB) {}


template<class T> template<class U> Range<T>::Range(U lower, U upper,
				  RangeBound lowerB, RangeBound upperB) :
  _lowerB(lowerB), _upperB(upperB) {
    
    _lower = lower;
    _upper = upper;
}


template<class T> void Range<T>::values(T &lower, T &upper,
			       RangeBound &lowerB, RangeBound &upperB) const{
  lower = _lower;
  upper = _upper;
  lowerB = _lowerB;
  upperB = _upperB;
}


template<class T> template<class U> bool 
Range<T>::operator<(const Range<U>& inRange) const{
  U lower, upper;
  RangeBound lowerB, upperB;
  
  inRange.values(lower, upper, lowerB, upperB);

  // First sort based on the lower limit.

  if( _lowerB!=Unbounded && lowerB!=Unbounded ){

    if( _lower < lower ) return true;
    if( _lower > lower)  return false;

    // A closed interval is taken to be less than an equivalent open interval.

    if( _lowerB==Closed && lowerB==Open ) return true;
    if( _lowerB==Open && lowerB==Closed ) return false;

  }else{

    // an undefined lower limit is less than a specified one.

    if( _lowerB==Unbounded && lowerB!=Unbounded ) return true;
    if( _lowerB!=Unbounded && lowerB==Unbounded ) return false;
  }

  // Lower limits are equivalent, so now sort on upper limits.

  if( _upperB!=Unbounded && upperB!=Unbounded ){

    if( _upper < upper ) return true;
    if( _upper > upper)  return false;

    // A closed interval is taken to be greater than an equivalent 
    // open interval.

    if( _upperB==Closed && upperB==Open ) return false;
    if( _upperB==Open && upperB==Closed ) return true;

  }else{

    // an undefined upper limit is greater than a specified one.

    if( _upperB==Unbounded && upperB!=Unbounded ) return false;
    if( _upperB!=Unbounded && upperB==Unbounded ) return true;
  }

  // The two ranges are identical

  return false;
}


template<class T> template<class U> bool 
Range<T>::operator==(const Range<U>& inRange) const{
  U lower, upper;
  RangeBound lowerB, upperB;
  bool lowMatch=true, highMatch=true;
  
  inRange.values(lower, upper, lowerB, upperB);

  if( lowerB!=_lowerB || upperB!=_upperB ) return false;

  if( lowerB!=Unbounded ) lowMatch = lower==_lower;

  if( upperB!=Unbounded ) highMatch = upper==_upper;

  return highMatch && lowMatch;
}


template<class T> template<class U> bool 
Range<T>::operator!=(const Range<U>& inRange) const{
  return !(*this==inRange);
}


template<class T>
list<Range<T> > parseRanges(const string &inString){
  list<Range<T> > outRanges;
  string parseString = inString;

  RangeBound upperB, lowerB;
  T upper = 0, lower = 0;

  while( parseString.size() != 0 ){
    string::size_type icomma = parseString.find(",");
    if( icomma == string::npos ) icomma = parseString.size();
    string rangeString = parseString.substr(0,icomma);
    parseString.erase(0,icomma+1);

    switch( *rangeString.substr(0,1).data() ){
    case '[':
      lowerB = Closed;
      rangeString.erase(0,1);
      break;
    case '(':
      lowerB = Open;
      rangeString.erase(0,1);
      break;
    case ':':
      lowerB = Unbounded;
      break;
    default :
      lowerB = Closed;
    }

    string::size_type iparse = rangeString.size();
    switch( *rangeString.substr(iparse-1,1).data() ){
    case ']':
      upperB = Closed;
      rangeString.erase(iparse-1,1);
      break;
    case ')':
      upperB = Open;
      rangeString.erase(iparse-1,1);
      break;
    case ':':
      upperB = Unbounded;
      break;
    default :
      upperB = Closed;
    }

    string::size_type icolon = rangeString.find(":");
    if( icolon==string::npos ){
      upper = lower = (T) atof(rangeString.c_str());
    }else{
      if( icolon!=0 )
	lower = (T) atof(rangeString.substr(0,icolon).c_str());
      if( icolon!=rangeString.size()-1 )
	upper = (T) atof(rangeString.substr(icolon+1).c_str());
    }

    Range<T> tempRange(lower, upper, lowerB, upperB);
    outRanges.push_back(tempRange);
  }

  return outRanges;
}

template<class T>
Range<T> maxRange(const list<Range<T> > &ranges){
  T upper, lower, upperMax, lowerMax;
  RangeBound upperB, lowerB, upperBmax, lowerBmax;

  list<Range<T> >::const_iterator irange;

  if( ranges.size()==0 ){
    Range<T> max(0, 0, Unbounded, Unbounded);
    return max;
  }

  ranges.begin()->values(lowerMax, upperMax, lowerBmax, upperBmax);

  for(irange=ranges.begin()++; irange!=ranges.end(); irange++){
    irange->values(lower, upper, lowerB, upperB);

    if( lowerB == Unbounded ) lowerBmax = Unbounded;
    else if( lower == lowerMax ){
      if( lowerB == Closed ) lowerBmax = Closed;
    }else if( lower < lowerMax ){
      lowerMax = lower;
      lowerBmax = lowerB;
    }

    if( upperB == Unbounded ) upperBmax = Unbounded;
    else if( upper == upperMax ){
      if( upperB == Closed ) upperBmax = Closed;
    }else if( upper > upperMax ){
      upperMax = upper;
      upperBmax = upperB;
    }
  }

  Range<T> max(lowerMax, upperMax, lowerBmax, upperBmax);
  return max;
}

template class Range<long>;
template class Range<double>;


template Range<long>::Range(const Range<double> &);
template Range<double>::Range(const Range<long> &);

template Range<double>::Range(int, int, RangeBound, RangeBound);
template Range<double>::Range(long, long, RangeBound, RangeBound);
template Range<double>::Range(unsigned, unsigned, RangeBound, RangeBound);
template Range<double>::Range(float, float, RangeBound, RangeBound);
template Range<double>::Range(short, short, RangeBound, RangeBound);
template Range<double>::Range(unsigned char, unsigned char, RangeBound, RangeBound);

template list<Range<long> > parseRanges<long>(const string &);
template list<Range<double> > parseRanges<double>(const string &);

template Range<long> maxRange(const list<Range<long> > &);
template Range<double> maxRange(const list<Range<double> > &);


template bool Range<long>::operator==(const Range<long>& inRange) const;
template bool Range<long>::operator==(const Range<double>& inRange) const;
template bool Range<double>::operator==(const Range<long>& inRange) const;
template bool Range<double>::operator==(const Range<double>& inRange) const;


template bool Range<long>::operator!=(const Range<long>& inRange) const;
template bool Range<long>::operator!=(const Range<double>& inRange) const;
template bool Range<double>::operator!=(const Range<long>& inRange) const;
template bool Range<double>::operator!=(const Range<double>& inRange) const;


template bool Range<long>::operator<(const Range<long>& inRange) const;
template bool Range<long>::operator<(const Range<double>& inRange) const;
template bool Range<double>::operator<(const Range<long>& inRange) const;
template bool Range<double>::operator<(const Range<double>& inRange) const;
